/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * CMDGADG.C
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 02/17/92
 *   MODIFIED : 07/26/94
 *       TABS : 04 08
 *
 * DESCRIPTION
 *		Command Button Gadget handling routines.
 *
 * HISTORY
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <echidna/etypes.h>
#include <echidna/hitareas.h>
#include <echidna/eerrors.h>
#include <echidna/eui.h>
#include <echidna/gadgets.h>
#include <echidna/hottext.h>
#include <echidna/cmdgadg.h>
#include <echidna/mouse.h>
#include <echidna/naming.h>


/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/


/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/


/*********************************************************************
 *
 * FitCGAD
 *
 * SYNOPSIS
 *		short FitCGAD (BTN *pbtn)
 *
 * PURPOSE
 *		To check if the text/shape will fit in the given button size.
 *		If size in pbtn->cx, pbtn->cy is too small it will fill in the 
 *		minimum size.
 *
 * INPUT
 *		pbtn: Pointer to initialized button struct. Fields cx and cy
 *			will be increased if necessary to fit mimimum size of button.
 *	
 * AFFECTS
 *		Fills in the fields 'cx' and 'cy' of pbtn with 
 *		minimum size if they are not big enough.
 *
 * RETURN VALUE
 *		TRUE if button can be drawn in requested dimension.
 *		FALSE if button must be enlarged to fit text/shape.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short FitCGAD (BTN *pbtn)
{

	short cxDecalWdth;
	short cyDecalHght;
	short cxDecalWdthMin;
	short cyDecalHghtMin;
	short cxMinWdth;
	short cyMinHght;
	short cxWdth = pbtn->cx;
	short cyHght = pbtn->cy;

	GetDimBDK (pbtn->bdk, pbtn->pvDecal, &cxDecalWdth, &cyDecalHght);

	cxDecalWdthMin = cxDecalWdth + cyCGADMin;
	cyDecalHghtMin = cyDecalHght + cxCGADMin;

	cxMinWdth = max (cxCGADMin, cxDecalWdthMin);
	cyMinHght = max (cyCGADMin, cyDecalHghtMin);

	pbtn->cx = max (cxWdth, cxMinWdth);
	pbtn->cy = max (cyHght, cyMinHght);
	
	return (cxWdth == pbtn->cx && cyHght == pbtn->cy); 

} /* FitCGAD */

/*********************************************************************
 *
 * CGADHandler
 *
 * SYNOPSIS
 *		static short CGADHandler (HitInfo *phi)
 *
 * PURPOSE
 *		Universal button handler.
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static short CGADHandler (HitInfo *phi)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CGADHandler";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
	HitArea *pha  = phi->HitArea;
	GHI		*pghi = (GHI *)pha->UserData;
	BTN		*pbtn = (BTN *)pghi->pvgad;
	UWORD	 fghiFlags = pghi->fhaFlags;
	UWORD	 fhaState = pha->fState;
	short	 Buttons = phi->Buttons;
	UBYTE	 fbtnNewState;
	
	fbtnNewState = 
		((phi->inpt == inptMouse && 
			(Buttons & (MOUSE_LEFTBUTTON | MOUSE_RIGHTBUTTON))) ? 
			BTNS_PRESSED : 0) |
		((fhaState & HA_SPACE) ? BTNS_MARQUEED : 0) |
		((fhaState & HA_ENTER) ? BTNS_HILIGHTED : 0) |
		((fhaState & HA_DISABLED) ? BTNS_GHOSTED : 0);



	if (fbtnNewState != pbtn->fState) {
		pbtn->fState = fbtnNewState;
		DrawCGAD (pbtn, pbtn->x, pbtn->y);
	} 

	/* Call specific button function if flags dictate a call is necessary */
	if (phi->inpt == inptHitKey || 
		(
			(phi->inpt == inptMouse) &&
			(
				((Buttons & MOUSE_FIRSTDOWN) && (fghiFlags & HTA_DOWN)) ||
				(!((Buttons & MOUSE_FIRSTDOWN) || !Buttons)) 	  	||
				((!Buttons) && !(fghiFlags & HTA_NO_UP))
			)
		)
	) {
		pha->UserData = pghi->pvData;
		pghi->pfngad(phi);
		pha->UserData = pghi;
	}
	return TRUE;
}} /* CGADHandler */


/*********************************************************************
 *
 * phaInitCGADHitArea
 *
 * SYNOPSIS
 *		HitArea *phaInitCGADHitArea (
 *			short x, short y,
 *			BTN		*pbtn,
 *			GHI		*pghi,
 *			PFNGAD	 pfngad,
 *			void	*pvData,
 *			HitArea *pha,
 *			UWORD	fhaFlags,
 *			UWORD	fhaState
 *		)
 *
 * PURPOSE
 *		To set up hit area for button and interpose standard button handler
 *		function between hit area routines and specific button function.
 *
 * INPUT
 *		x,y:	Coordinates of button/hit area (to go in hitarea X and Y).
 *		pbtn:	Pointer to Initialized Button struct (except 'fState' field).
 *		pghi:	Pointer to Uninitialized Button Area Info struct.
 *		pfngad: Pointer to specific button function (non-NULL).
 *		pvData: Pointer to Initialized Data for specific button function.
 *		pha:	Pointer to Uninitialized HitArea struct (except 'ID' field).
 *		fhaFlags:	Flags for hit area. Acceptable flags:
 *					HTA_DOWN, HTA_REPEAT, HTA_INSURE_UP, HTA_NO_UP, HTA_ESCAPE, 
 *					HTA_DEFAULT_ENTER, HTA_SPACE_FOCUS
 *					HTA_ENTER_FOCUS.
 *		fhaState: HitArea dynamic state flags.  Acceptable flags:
 *					HA_SPACE, HA_ENTER, HA_DISABLED
 *		
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		Returns input parameter 'pha'.
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
HitArea *phaInitCGADHitArea (
	short x, short y,
	BTN		*pbtn,
	GHI		*pghi,
	PFNGAD	 pfngad,
	void	*pvData,
	HitArea *pha,
	UWORD	fhaFlags,
	UWORD	fhaState
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "phaInitCGADHitArea";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
	pbtn->fState = 
		((fhaState & HA_SPACE) ? BTNS_MARQUEED : 0) |
		((fhaState & HA_ENTER) ? BTNS_HILIGHTED : 0) |
		((fhaState & HA_DISABLED) ? BTNS_GHOSTED : 0);
	pbtn->rlpsDecal = rlpsOn;
	pbtn->jstfDecal = jstfCenter;

	pghi->pvgad = pbtn;
	pghi->pfngad = pfngad;
	pghi->pvData = pvData;
	pghi->fhaFlags = fhaFlags & (HTA_DOWN | HTA_INSURE_UP | HTA_NO_UP | HTA_REPEAT);

	pha->X = x;
	pha->Y = y;
	pha->Width = pbtn->cx;
	pha->Height = pbtn->cy;
	pha->Flags = (fhaFlags | HTA_DOWN | HTA_INSURE_UP) &
			(HTA_DOWN | HTA_REPEAT | HTA_INSURE_UP | HTA_ESCAPE |
			HTA_DEFAULT_ENTER | HTA_ENTER_FOCUS | HTA_SPACE_FOCUS);

	/* pha->ID pre-initialized by caller */
	pha->Data = CGADHandler;
	pha->UserData = pghi;
	pha->KeyCode = 0;
	if (pbtn->bdk == bdkHotText) {
		char *pszHotText = (char *)pbtn->pvDecal;
		pha->KeyCode = GetHotTextHotKey (pszHotText);
	}
	pha->fState = fhaState & (HA_SPACE | HA_ENTER | HA_DISABLED);

	return pha;
	
}} /* phaInitCGADHitArea */

