/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * GADGETS.C
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 03/06/92
 *   MODIFIED : 02/05/95
 *       TABS : 04 08
 *
 *	     \|///-_
 *	     \oO///_
 *	-----w/-w-----
 *	 E C H i D N A
 *	--------------
 *
 * DESCRIPTION
 *		Low level gadget handling functions. Having mostly to do with
 *		drawing of various kinds of buttons.
 *
 * HISTORY
 *		02/05/95 (dcc) - Comment out calls to SetRectPat() to silence warnings.
 *						 (All are marked with '// KLUDGE ')
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <echidna/etypes.h>
#include <echidna/grafx.h>
#include <echidna/gadgets.h>
#include <echidna/HotText.h>
#include <echidna/scic.h>
#include <echidna/eui.h>
#include <echidna/windows.h>
#include <echidna/circle.h>
#include <echidna/eerrors.h>
#include <echidna/naming.h>

/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/


/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/


/*********************************************************************
 *
 * GetDimBDK
 *
 * SYNOPSIS
 *		void GetDimBDK (BDK bdk, void *pvDecal, short *cx, short *cy)
 *
 * PURPOSE
 *		Returns dimensions of decal.
 *
 * INPUT
 *		bdk:		Button decal kind.
 *		pvDecal:	Decal data pointer.
 *		cx, cy:		Pointer to dimension variables to be filled in (non-NULL).
 *
 * AFFECTS
 *		Fills in *cx and *cy with dimensions of decal when drawn, given 
 *		current active window.
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void GetDimBDK (BDK bdk, void *pvDecal, short *cx, short *cy)
{
	switch (bdk) {
	case bdkHotText:	/* Text Decal */
		{ 
			Font *pfont = GetActiveWindow()->Font;

			*cx = HotTextLen((char *)pvDecal);
			*cy = pfont->Letters[0]->Height;
		}
		break;
	case bdkSCIC:	/* Scaled Icon Decal */
		{
			*cx = ((SCIC *)pvDecal)->cx;
			*cy = ((SCIC *)pvDecal)->cy;
		}
		break;
	case bdkGShape:	/* GShape Decal */
		{ 
			*cx = ((GSH *)pvDecal)->Width;
			*cy = ((GSH *)pvDecal)->Height;
		}
		break;
	case bdkBlank:
		{ 
			*cx = *cy = 0;
		}
		break;
	}
	
} /* GetDimBDK */

/*********************************************************************
 *
 * DrawBDK
 *
 * SYNOPSIS
 *		void DrawBDK (
 *			BDK bdk, 
 *			void *pvDecal, 
 *			short x, short y,
 *			UBYTE fState,
 *			UBYTE fPref
 *		)
 *
 * PURPOSE
 *		To draw the decal according to bdk and fState, at (x,y). In current
 *		Pen (foreground) and Paper (background) colors.
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DrawBDK (
	BDK bdk, 
	void *pvDecal, 
	short x, short y, 
	UBYTE fState, 
	UBYTE fPref
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DrawBDK";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
	short pencolor = GetPenColor();
	short cx, cy;

	/* Get decal dimensions */
	GetDimBDK (bdk, pvDecal, &cx, &cy);

	if (cx && cy) {
		BeforeGraphics ();
		/* Clear decal area and Marquee area */
		SetPenColor (GetPaperColor());
		DrawRect (x-2, y-2, cx + 4, cy + 4);

		SetPenColor (pencolor);
		switch (bdk) {
		case bdkHotText:
			{
				DrawHotText (x, y, (char *)pvDecal);
			}
			break;
		case bdkSCIC:
			{
				(((SCIC *)pvDecal)->pfnscic)(x, y,(SCIC *)pvDecal);
			}
			break;
		case bdkGShape:
			{
				DrawGShape ((GSH *)pvDecal, x, y);
			}
			break;
		}

		if (fState & (BTNS_MARQUEED | BTNS_GHOSTED)) {
// KLUDGE 			SetRectPat (rgpubEUIGhostPats[fPref & BTNP_ODD_GHOST]);
			fPref = fPref;	// KLUDGE 
			if (fState & BTNS_MARQUEED) {
				DrawPatBox (x - 2, y - 2, cx + 4, cy + 4);
			}
			if (fState & BTNS_GHOSTED) {
				SetPenColor (GetPaperColor());
				DrawPatRect (x, y, cx, cy);
			}
		}

		AfterGraphics ();
	}
}} /* DrawBDK */

/*********************************************************************
 *
 * GetRelativePositionDeltas
 *
 * SYNOPSIS
 *		void GetRelativePositionDeltas (
 *			short cxA, short cyA,
 *			short cxB, short cyB,
 *			RLPS rlps, JSTF jstf,
 *			short *pdxA, short *pdyA
 *		)
 *
 * PURPOSE
 *		To get the position delta of a decal relative to a button.
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void GetRelativePositionDeltas (
	short cxA, short cyA,
	short cxB, short cyB,
	RLPS rlps, JSTF jstf,
	short *pdxA, short *pdyA
)
{
	switch (rlps) {
	case rlpsRight: case rlpsLeft:
		if (rlps == rlpsRight) {
			*pdxA = cxB + cxB / 2;
		} else {
			*pdxA = -cxA - cxB / 2;
		}
		*pdyA = 0;
		switch (jstf) {
		case jstfRight:
			*pdyA = cyB;
			break;
		case jstfCenter:
			*pdyA = (cyB - cyA) / 2;
			break;
		}
		break;
	case rlpsBelow: case rlpsAbove:
		if (rlps == rlpsBelow) {
			*pdyA = cyB + cyB / 4;
		} else {
			*pdyA = -cyA - cyB / 4;
		}
		*pdxA = 0;
		switch (jstf) {
		case jstfRight:
			*pdxA = cxB - cxA;
			break;
		case jstfCenter:
			*pdxA = (cxB - cxA) / 2;
			break;
		}
		break;
	default:	/* On */
		*pdxA = (cxB - cxA) / 2;
		*pdyA = (cyB - cyA) / 2;
		break;
	}

} /* GetRelativePositionDeltas */

/*********************************************************************
 *
 * DrawPushBTN
 *
 * SYNOPSIS
 *		void DrawPushBTN (
 *			short x, short y,
 *			short cx, short cy,
 *			UBYTE fState,
 *			UBYTE fPref
 *		)
 *
 * PURPOSE
 *		Draw a key like button that looks like this:
 *	
 *     Normal           HighLighted        Pressed     Pressed & Highlighted
 *	
 *  ***************   ***************   ***************   ***************       
 * *              &* ***************** *&&&&&&&&&&&&&&&* *****************
 * *             &&* **            &** *&&&&&&&&&&&&&&** **&&&&&&&&&&&&&**
 * *  %%%%%%%%%%%&&* **           &&** *&&%%%%%%%%%%%%** **&&&&&&&&&&&&&**
 * *  %%%%%%%%%%%&&* **  %%%%%%%%%&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * *  %%%%%%%%%%%&&* **  %%%%%%%%%&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * *  %%%%%%%%%%%&&* **  %%%%%%%%%&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * *  %%%%%%%%%%%&&* **  %%%%%%%%%&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * *  %%%%%%%%%%%&&* **  %%%%%%%%%&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * *  %%%%%%%%%%%&&* **  %%%%%%%%%&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * *  %%%%%%%%%%%&&* **  %%%%%%%%%&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * *  %%%%%%%%%%%&&* ** &&&&&&&&&&&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * * &&&&&&&&&&&&&&* **&&&&&&&&&&&&&** *&&%%%%%%%%%%%%** **&&%%%%%%%%%%%**
 * *&&&&&&&&&&&&&&&* ***************** ***************** *****************
 *  ***************   ***************   ***************   *************** 
 *
 * INPUT
 *		x,y:	Coordinate of upper left corner.
 *		cx,cy:	Dimensions.
 *		fState:	BTNS_MARQUEED | BTNS_HILIGHTED | BTNS_PRESSED | BTNS_GHOSTED
 *		fPref:	BTNP_ODD_GHOST.
 *	
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DrawPushBTN (
	short x, short y,
	short cx, short cy,
	UBYTE fState,
	UBYTE fPref
)
{
	Window	*pwin = GetActiveWindow();
	short BevelLight, BevelDark;


	BeforeGraphics ();

	/* Black Outline Box with rounded corners.*/
	SetPenColor (pwin->Black);
	DrawRect (x+1, y, cx-2, 1);				/* top */
	DrawRect (x+1, y+cy-1, cx-2, 1);	/* bottom */
	DrawRect (x, y+1, 1, cy-2);				/* left */
	DrawRect (x+cx-1, y+1, 1, cy-2);	/* right */

	/* Determine beveled edge colors */
	if (fState & BTNS_PRESSED) {
		BevelLight = pwin->DarkGray;
		BevelDark = pwin->LightGray;
	/*	fState |= BTNS_HILIGHTED;*/
	} else {
		BevelLight = pwin->White;
		BevelDark = pwin->DarkGray;
	}

	if (fState & BTNS_HILIGHTED) {
		cx -= 2;
		cy -= 2;
		++x;
		++y;
		DrawBox (x, y, cx, cy);
	}

	/* Beveled Edges */
	SetPenColor (BevelLight);
	DrawRect (x+1, y+1, cx-2, 2);	/* white top */
	DrawRect (x+1, y+3, 2, cy-4);	/* white left side */

	SetPenColor (BevelDark);
	if (fState & BTNS_PRESSED) {
		DrawRect (x+3, y+cy-3, cx-4, 2); /* dark grey Bottom */
		DrawRect (x+cx-3, y+3, 2, cy-6); /* dark gray right side */
	} else {
		DrawRect (x+2, y+cy-3, cx-3, 2); /* dark grey Bottom */
		DrawRect (x+cx-3, y+2, 2, cy-5); /* dark gray right side */
		DrawPixel (x+1, y+cy-2);			/* lower left corner pixel */
		DrawPixel (x+cx-2, y+1);			/* upper right corner pixel*/	
	}

	/* Surface */
	SetPenColor (pwin->LightGray);
	DrawRect (x+3, y+3, cx-6, cy-6);

	
	/* Set up pen colors for decal drawing */
	SetPenColor (pwin->Black);
	SetPaperColor (pwin->LightGray);

	AfterGraphics ();

	fPref = fPref;

} /* DrawPushBTN */

/*********************************************************************
 *
 * DrawBullsEyeBTN
 *
 * SYNOPSIS
 *		void DrawBullsEyeBTN (
 *			short x, short y, 
 *			short cx, short cy, 
 *			UBYTE fState,
 *			UBYTE fPref
 *		)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DrawBullsEyeBTN (
	short x, short y, 
	short cx, short cy, 
	UBYTE fState,
	UBYTE fPref
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DrawBullsEyeBTN";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
	short radius = RadiusFromDiameter(cy);
	short xCenter = x + radius;
	short yCenter = y + radius;

	BeforeGraphics ();

	SetPenColor (GetActiveWindow()->White);
	DrawRect (x, y, cx, cy);

	SetPenColor (GetActiveWindow()->Black);
	DrawCircle (xCenter, yCenter, radius);

	if (fState & BTNS_HILIGHTED) {
		DrawCircle (xCenter, yCenter, radius - 1);
	}

	if (fState & BTNS_PRESSED) {
		DrawFCircle (xCenter, yCenter, radius/2);
	}

	if ((fState & BTNS_GHOSTED) && cx && cy) {
		SetPenColor (GetActiveWindow()->White);
// KLUDGE 		SetRectPat (rgpubEUIGhostPats[fPref & BTNP_ODD_GHOST]);
		fPref = fPref;	// KLUDGE 
		DrawPatRect (x, y, cx, cy);
	}

	/* Set up pen colors for decal drawing */
	SetPenColor (GetActiveWindow()->Black);
	SetPaperColor (GetActiveWindow()->White);

	AfterGraphics ();

}} /* DrawBullsEyeBTN */


/*********************************************************************
 *
 * DrawCheckBoxBTN
 *
 * SYNOPSIS
 *		void DrawCheckBoxBTN (
 *			short x, short y,
 *			short cx, short cy,
 *			UBYTE fState,
 *			UBYTE fPref
 *		)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DrawCheckBoxBTN (
	short x, short y,
	short cx, short cy,
	UBYTE fState,
	UBYTE fPref
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DrawCheckBoxBTN";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
	if (!(cx & 1)) ++cx;
	if (!(cy & 1)) ++cy;

	BeforeGraphics ();
	
	SetPenColor (GetActiveWindow()->White);
	DrawRect (x, y, cx, cy);

	SetPenColor (GetActiveWindow()->Black);
	DrawBox (x, y, cx, cy);

	if (fState & BTNS_HILIGHTED) {
		DrawBox (x+1, y+1, cx - 2, cy - 2);
	}

	if (fState & BTNS_PRESSED) {
		short i;
		
		for (i = 1; i < (cx / 2); i++) {
			DrawPixel (x + i, y + i);
			DrawPixel (x + cx - 1 - i, y + i);
			DrawPixel (x + i, y + cy - 1 - i);
			DrawPixel (x + cx - 1 - i, y + cy - 1 - i);
		}
		DrawPixel (x + i, y + i);
	}

	if (fState & BTNS_GHOSTED) {
		SetPenColor (GetActiveWindow()->White);
// KLUDGE 		SetRectPat (rgpubEUIGhostPats[fPref & BTNP_ODD_GHOST]);
		fPref = fPref;	// KLUDGE 
		DrawPatRect (x, y, cx, cy);
	}

	/* Set up pen colors for decal drawing */
	SetPenColor (GetActiveWindow()->Black);
	SetPaperColor (GetActiveWindow()->White);

	AfterGraphics ();
}} /* DrawCheckBoxBTN */


typedef void (* PFNDRAWBTN)(
	short x, short y, 
	short cx, short cy, 
	UBYTE fState,
	UBYTE fPref
);
#define cbtnkMax	3
static PFNDRAWBTN rgpfnDrawBTNK[cbtnkMax] = {
	DrawCheckBoxBTN,
	DrawBullsEyeBTN,
	DrawPushBTN,
};

/*********************************************************************
 *
 * DrawBTN
 *
 * SYNOPSIS
 *		void DrawBTN (BTN *pbtn, short x, short y)
 *
 * PURPOSE
 *		Draw a button.
 *
 * INPUT
 *		pbtn:	Button pointer.
 *		x,y:	Coodinates.
 *	
 * AFFECTS
 *		Fills in pbtn->x and pbtn->y with x and y, respectively.
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DrawBTN (BTN *pbtn, short x, short y)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DrawBTN";
#endif
#if REQUIRE
	if (!pbtn || pbtn->btnk < 0 || pbtn->btnk > cbtnkMax) {
		SetGlobalErr (ERR_REQUIREMENTS);
		GEprintf ("DrawBTN\n");
	}
#endif
/*-------------------------------------------------------------------------*/
{
	short cxDecal, cyDecal;
	short cxButton = pbtn->cx;
	short cyButton = pbtn->cy;
	UBYTE fState = pbtn->fState;
	UBYTE fPref = pbtn->fPref;
	short dxDecal, dyDecal;

	pbtn->x = x;
	pbtn->y = y;

	/* Draw appropriate button */
	rgpfnDrawBTNK[pbtn->btnk](x, y, cxButton, cyButton, fState, fPref);

	/* Get decal dimensions */
	GetDimBDK (pbtn->bdk, pbtn->pvDecal, &cxDecal, &cyDecal);

	GetRelativePositionDeltas (
		cxDecal, cyDecal,
		cxButton, cyButton, 
		pbtn->rlpsDecal,
		pbtn->jstfDecal,
		&dxDecal, &dyDecal
	);

	if (pbtn->btnk == btnkPush) {
		if ((pbtn->rlpsDecal & rlpsOn) && (fState & BTNS_PRESSED)) {
			++dxDecal;
			++dyDecal;
		}
		if (pbtn->rlpsDecal != rlpsOn) {
			SetPaperColor (GetActiveWindow()->White);
		}
	}


	/* Draw decal in appropriate place */
	DrawBDK (pbtn->bdk, pbtn->pvDecal, x + dxDecal,	y + dyDecal, fState, fPref);

}} /* DrawBTN */

