/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * EMS.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 01/15/91
 *   MODIFIED : 10/08/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		EMS routines
 *
 * HISTORY
 *		03/25/92 (dcc) - remove repeated calls to DeleteEMSPntr()
 *						 with same EMSPntr in SpecialFreeEMS().
 *		05/06/92 (dcc) - when all EMS pages were assigned by the
 *						 user, ReadEMS() and WriteEMS() did not
 *						 properly restore user's logical page mapping
 *		05/08/92 (dcc) - enhance ActivateEMSatWin() so if page is
 *						 already mapped in, it won't do it again
 *		08/29/92 (GAT) - Added support for GMEMLIB and it's
 *						 restrictions on freeing memory. Books
 *						 are now freed when they are no longer
 *						 in use and if GMEMLIB then the XMSPntr
 *						 mechanism is compiled out.
 *		11/18/92 (GAT) - changed AvailEMS() to return -1 if
 *						 not used.
 *		01/14/93 (dcc) - bug fix: when books are freed, we must
 *						 indicate that all pages with the same handle
 *						 are no longer valid.
 *		03/07/94 (dcc) - Add SetEMSMemoryPoolSize()
 *		10/06/94 (dcc) - updated to use <echidna/alloc.h>
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <echidna/etypes.h>
#include <echidna/alloc.h>
#include <echidna/eerrors.h>
#include <echidna/listfunc.h>
#include <echidna/exit.h>
#include <echidna/ems.h>

#include <dos.h>
#include <string.h>

#define LOCAL static

/**************************** C O N S T A N T S ***************************/

#define MAX_BOOK_PAGES	32
#define PAGE_SIZE		16384U

#define BUNCHOFEMSPNTR	74

/******************************** T Y P E S *******************************/

typedef struct {
	NodeType	 Node;
	UWORD		 EMMHandle;
	UWORD		 NumPages;
	ListType	 PageListsX[MAX_BOOK_PAGES];
	ListType	*PageLists[MAX_BOOK_PAGES];
} Book;

typedef struct {
	NodeType	 Node;
	EMSPntr		*EMS;
} EMSTracker;

typedef struct {
	NodeType	 Node;
	UWORD		 Offset;
	UWORD		 Size;
} FreeMemType;

/****************************** G L O B A L S *****************************/

static UWORD cEMSMaxPages = 0;			/* How much EMS we may allocate */
static UWORD cEMSPagesAllocated = 0;	/* # EMS pages already allocated */

/*	sizeofEMSPntr - defaults to sizeof (EMSPntr), which means you can
	allocate about 65535 bytes. Other values set by SetLargestEMS:
	sizeof (EMSPntr16384) <==> allocate blocks up to 16384 bytes,
	sizeof (EMSPntr32768) <==> allocate blocks up to 32768 bytes, and
	sizeof (EMSPntr49152) <==> allocate blocks up to 49152 bytes. */

static UWORD sizeofEMSPntr = sizeof (EMSPntr);

LOCAL ListType		 BookListX;
LOCAL ListType		*BookList = &BookListX;
LOCAL ListType		 EMSStackX;
LOCAL ListType		*EMSStack = &EMSStackX;
LOCAL ListType		 EMSFreeListX;
LOCAL ListType		*EMSFreeList = &EMSFreeListX;
LOCAL UWORD			 PageFrame = 0;
LOCAL UBYTE			*PageFramePtr;
LOCAL EMSPntr		*LastEMS;

/*	The following variables are used by ReadEMS() and WriteEMS() to determine
	who mapped a particular physical EMS page. FindAndMapIntoPhysicalPage()
	first checks to see if a logical EMS has already been mapped into a
	physical page; if it finds it, it uses it. Else it looks for a free
	physical page (not yet mapped, or mapped by EMSMAPS), and uses that
	physical page to deal with EMS memory. If no free physical pages are
	available, it save the current page 0 logical mapping, does all its
	I/O thru page 0, and restores the page 0 logical mapping. */

#define EMSMAPS		1
#define USERMAPS	2

static UWORD uwPageEMMHandle[EMS_PAGES];	/* Save last handle in EMS physical page. */
static UBYTE ubPageLogicalPage[EMS_PAGES];	/* Save last logical page in EMS physical page. */
static BOOL  fMappedBy[EMS_PAGES];			/* Who mapped EMS physical page? */

/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/


/*********************************************************************
 *
 * SetEMSMemoryPoolSize
 *
 * PURPOSE
 *		Call this routine to set the maximum amount of EMS (in Kbytes)
 *		that may be allocated via AllocEMS().
 *
 *		It is NOT necessary to call this routine. It is probably
 *		most useful to assist in debugging low EMS memory conditions.
 *
 *		If you call this routine with 0, it will use all available EMS.
 *
 * INPUT
 *		UWORD cEMSKBytes	: number of K bytes to limit EMS memory
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *		03/07/94 Monday (dcc) - created.
 *
*/
void SetEMSMemoryPoolSize(UWORD cEMSKBytes)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "SetEMSMemoryPoolSize";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	cEMSMaxPages = cEMSKBytes / 16;

} /* SetEMSMemoryPoolSize */


/*********************************************************************
 *
 * SetLargestEMS
 *
 * SYNOPSIS
 *		void SetLargestEMS (size_t size)
 *
 * PURPOSE
 *		Defines the largest contiguous block the program will ever
 *		allocate. Typically, you will call this routine with a multiple
 *		of 16384. The intent is that if you specify the maximum to
 *		be less than 64K, the EMS routines may be able to save memory
 *		in allocating the pointers to keep track of an EMS memory block.
 *		If this routine is never called, the value defaults to the
 *		pre-initialized value szLargestXTRA.
 *
 * INPUT
 *		size - the largest block your program will ever allocate.
 *
 * EFFECTS
 *		if size <= 16384, EMSPntr will include space for only one EMMTracker.
 *		if size <= 32768, EMSPntr will include space for two EMMTrackers.
 *		if size <= 49152, EMSPntr will include space for three EMMTrackers.
 *		otherwise (& default), EMSPntr will include space for four EMMTrackers.
 *
 * RETURN VALUE
 *		NONE.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void SetLargestEMS (size_t size)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "SetLargestEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (size <= 16384U)
	{
		sizeofEMSPntr = sizeof (EMSPntr16384);
	}
	else if (size <= 32768U)
	{
		sizeofEMSPntr = sizeof (EMSPntr32768);
	}
	else if (size <= 49152U)
	{
		sizeofEMSPntr = sizeof (EMSPntr49152);
	}

} /* SetLargestEMS */


/*********************************************************************
 *
 * NewEMSPntr
 *
 * SYNOPSIS
 *		LOCAL EMSPntr *NewEMSPntr (void)
 *
 * PURPOSE
 *		Return a EMSPntr from EMSFree if available,
 *		otherwise, allocate a new block of EMSPntr.
 *
 *		Rationale: currently, sizeof (EMSPntr) == 24. Turbo C allocates
 *		in 16 byte units. Therefore, if you allocate EMSPntr one at a
 *		time, you will waste 8 bytes each time (32 - 24 = 8). Thus we
 *		allocate BUNCHOFEMSPNTR at once.
 *
 *		Note: even though a EMSPntr does not have a node header, we still
 *		add it to EMSFreeList. There is an implied type conversion from
 *		struct EMSPntr to struct NodeType.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Pointer to a free EMSPntr to use.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL EMSPntr *NewEMSPntr (void)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "NewEMSPntr";
#endif
#if REQUIRE

#endif

/*-------------------------------------------------------------------------*/
	#if USE_GAMEMEM
	return calloc (1, sizeofEMSPntr);
	#else
	{
		UBYTE		*pMem;

		if ((pMem = RemTail(EMSFreeList)) != NULL)
		{
			memset(pMem, 0, sizeofEMSPntr);	/* zero out (crap) Node info */
			return (EMSPntr *) pMem;
		}
		else
		{
			pMem = MEM_calloc (BUNCHOFEMSPNTR, sizeofEMSPntr);
			if (pMem)
			{
				int i;

				for (i = 1; i < BUNCHOFEMSPNTR; i++)
				{
					AddTail(EMSFreeList, pMem + i * sizeofEMSPntr);
				}
				return (EMSPntr *) pMem;
			}
		}

		return NULL;
	}
	#endif

} /* NewEMSPntr */


/*********************************************************************
 *
 * DeleteEMSPntr
 *
 * SYNOPSIS
 *		LOCAL void DeleteEMSPntr (EMSPntr *pEMS)
 *
 * PURPOSE
 *		Add pEMS to EMSFreeList.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL void DeleteEMSPntr (EMSPntr *pEMS)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DeleteEMSPntr";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	#if USE_GAMEMEM
	free (pEMS);
	#else
	((NodeType *)(pEMS))->Type = 0;	/* Fix so AddTail does generate error */
	AddTail(EMSFreeList, pEMS);
	#endif

} /* DeleteEMSPntr */


/*********************************************************************
 *
 * FreeBook
 *
 * PURPOSE
 *		Free previously allocated EMS "book".
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		03/07/94 Monday (dcc) - add support for cEMSPagesAllocated
 *
 * SEE ALSO
 *
*/
void FreeBook (Book *bk)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FreeBook";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		union REGS	 inregs;
		union REGS	 outregs;
		short		 numpages;
		short		 i;

		numpages = bk->NumPages;
		if (numpages)
		{
			for (i = 0; i < numpages; i++)
			{
				if (bk->PageLists[i])
				{
					EmptyList (bk->PageLists[i]);
				}
			}
			inregs.h.ah = 0x45;
			inregs.x.dx = bk->EMMHandle;
			int86 (0x67, &inregs, &outregs);

			/* Make sure any pages in the frame are invalidated. */

			for (i = 0; i < EMS_PAGES; i++)
			{
				if (uwPageEMMHandle[i] == bk->EMMHandle)
				{
					uwPageEMMHandle[i] = ubPageLogicalPage[i] = fMappedBy[i] = 0;
				}
			}
			cEMSPagesAllocated -= numpages;
		}
		DeleteNode (bk);
	}
} /* FreeBook */


/*********************************************************************
 *
 * FreeUnusedBooks
 *
 * SYNOPSIS
 *		LOCAL void FreeUnusedBooks (void)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL void FreeUnusedBooks (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FreeUnusedBooks";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
	Book	*bk;

	bk = Tail (BookList);
	while (!StartOfList (bk))
	{
		Book		*prev;
		UWORD		 freepages = 0;
		UWORD		 page;

		prev = Prev (bk);

		for (page = 0; page < bk->NumPages; page++)
		{
			FreeMemType	*fmt;

			fmt = Head (bk->PageLists[page]);
			if (EndOfList (fmt))
			{
				break;
			}
			if (fmt->Size != PAGE_SIZE)
			{
				break;
			}
			freepages++;
		}

		if (freepages == bk->NumPages)
		{
			Remove (bk);
			FreeBook (bk);
		}
		
		bk = prev;
	}
}
} /* FreeUnusedBooks */


/*********************************************************************
 *
 * AddNewBook
 *
 * PURPOSE
 *		Add a new "Book" of EMS pages.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Returns TRUE if successful.
 *
 * HISTORY
 *		03/07/94 Monday (dcc) - add support for cEMSPagesAllocated
 *
 * SEE ALSO
 *
*/
LOCAL short AddNewBook (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "AddNewBook";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		union REGS	 inregs;
		union REGS	 outregs;
		Book		*bk;
		short		 i;
		short		 numpages;
		FreeMemType	*fmt;

		bk = CreateNode (sizeof (Book), NULL);
		if (bk)
		{
			inregs.h.ah = 0x42;
			int86 (0x67, &inregs, &outregs);

			if (outregs.h.ah)
			{
				SetGlobalErr (ERR_GENERIC);
				GEprintf1 ("Error (1) %02x in EMS driver", outregs.h.ah);
				goto anbcleanup;
			}

			numpages = min (MAX_BOOK_PAGES, outregs.x.bx);

			/* See if there is a limit on how much EMS we may allocate */

			if (cEMSMaxPages && cEMSMaxPages >= cEMSPagesAllocated)
				numpages = min(numpages, cEMSMaxPages - cEMSPagesAllocated);

			if (!numpages)
			{
				goto anbcleanup;
			}

			inregs.h.ah = 0x43;
			inregs.x.bx = numpages;
			int86 (0x67, &inregs, &outregs);

			if (outregs.h.ah)
			{
				SetGlobalErr (ERR_GENERIC);
				GEprintf1 ("Error (2) %02x in EMS driver", outregs.h.ah);
				goto anbcleanup;
			}
			cEMSPagesAllocated += numpages;

			bk->EMMHandle = outregs.x.dx;
			bk->NumPages  = numpages;

			for (i = 0; i < numpages; i++)
			{
				bk->PageLists[i] = &bk->PageListsX[i];
				InitList (bk->PageLists[i]);
				fmt = CreateNode (sizeof (FreeMemType), NULL);
				if (!fmt)
				{
					goto anbcleanup;
				}
				fmt->Size = PAGE_SIZE;
				AddTail (bk->PageLists[i], fmt);
			}
			AddTail (BookList, bk);
			return TRUE;
		}
anbcleanup:
		if (bk)		FreeBook (bk);
		return FALSE;
	}

} /* AddNewBook */


/*********************************************************************
 *
 * FreeBooks
 *
 * SYNOPSIS
 *		void FreeBooks (void)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL void FreeBooks (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FreeBooks";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		Book		*bk;

		while ((bk = RemTail (BookList)) != NULL) {
			FreeBook (bk);
		}
	}

} /* FreeBooks */


/*********************************************************************
 *
 * InitEMS
 *
 * SYNOPSIS
 *		short InitEMS (void)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		10/08/94 (dcc) - add NULLTRAPOFF & NULLTRAPON to disable null-ptr trap
 *
 * SEE ALSO
 *
*/
LOCAL short InitEMS (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "InitEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		union REGS	 inregs;
		union REGS	 outregs;
		UWORD far32 	*w;
		char far32	*emsname;

		w = MK_FP (0x0000, 0x67 * 4 + 2);
		/* NULLTRAPOFF */
		emsname = MK_FP (*w, 10);

		if (strncmp ("EMMXXXX0", emsname, 8)) {
			/* EMS is NOT here */
			return FALSE;
		}
		/* NULLTRAPON */

		inregs.h.ah = 0x40;
		int86 (0x67, &inregs, &outregs);
		if (outregs.h.ah) {
			/* EMS is NOT functional */
			return FALSE;
		}

		inregs.h.ah = 0x41;
		int86 (0x67, &inregs, &outregs);
		if (outregs.h.ah) {
			/* EMS error */
			return FALSE;
		}
		PageFrame    = outregs.x.bx;
		PageFramePtr = MK_FP (PageFrame, 0);

		InitList (BookList);
		InitList (EMSFreeList);
		if (!AddNewBook ()) {
			/* EMS not allocated */
			return FALSE;
		}
		AddExitFunc (FreeBooks);
	}
	return TRUE;

} /* InitEMS */


/*********************************************************************
 *
 * MapPhysicalPage
 *
 * SYNOPSIS
 *		void MapPhysicalPage(UBYTE ubPhysicalPage,
 *			UWORD uwHandle, UBYTE ubLogicalPage)
 *
 * PURPOSE
 *		Map logical EMS page <uwHandle>, <ubPhysicalPage> to physical
 *		EMS page <ubPhysicalPage>.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static void MapPhysicalPage(UBYTE ubPhysicalPage, UWORD uwHandle, UBYTE ubLogicalPage)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "MapPhysicalPage";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		union REGS	 inregs;
		union REGS	 outregs;

		inregs.h.ah = 0x44;
		inregs.h.al = ubPhysicalPage;
		inregs.x.bx = ubLogicalPage;
		inregs.x.dx = uwHandle;
		int86 (0x67, &inregs, &outregs);
	}

} /* MapPhysicalPage */


/*********************************************************************
 *
 * FindPhysicalPage
 *
 * SYNOPSIS
 *		static int FindPhysicalPage(UBYTE *pMem)
 *
 * PURPOSE
 *		See if <pMem> falls in one of the physical EMS pages.
 *		If so, return the physical EMS page number, else return -1.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		The physical EMS page number <pMem> falls in, or -1.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static int FindPhysicalPage(UBYTE *pMem)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FindPhysicalPage";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if ((UBYTE huge32 *) PageFramePtr <= (UBYTE huge32 *) pMem &&
		 (UBYTE huge32 *) pMem <= (UBYTE huge32 *) PageFramePtr+0xFFFFU)
	{
		return (int) (((UBYTE huge32 *) pMem - PageFramePtr) / PAGE_SIZE);
	}

	return -1;

} /* FindPhysicalPage */


/*********************************************************************
 *
 * FindAndMapIntoPhysicalPage
 *
 * SYNOPSIS
 *		static UBYTE FindAndMapIntoPhysicalPage(
 *		 					UWORD uwHandle, UBYTE ubPage, UBYTE *pMem)
 *
 * PURPOSE
 *		Find a physical page to map the logical page into.
 *		1.	See if the logical page is already mapped into a physical
 *			page. If it is, just return the physical page number.
 *		2.	Else look for a free physical page (not yet mapped, or
 *			mapped by EMSMAPS), and map the logical page to that
 *			physical page.
 *		3.	Else map the logical page to a physical page not occupied
 *			by <pMem>..<pMem>+16383.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Physical EMS page mapped into.
 *
 * HISTORY
 *		04/23/92 Thursday (dcc) - enhanced with new argument <pMem>.
 *				If <pMem> resides in a physical EMS page, then this
 *				routine will not use that page for its I/O.
 *
 * SEE ALSO
 *
*/
static UBYTE FindAndMapIntoPhysicalPage(UWORD uwHandle, UBYTE ubPage, UBYTE *pMem)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FindAndMapIntoPhysicalPage";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		int i;

		/* See if logical page to read from is already mapped
			into a physical page. */

		for (i = 0; i < EMS_PAGES; i++)
		{
			if (fMappedBy[i] &&						/* If valid mapping, and... */
				uwPageEMMHandle[i] == uwHandle &&	/* same handle, and... */
				ubPageLogicalPage[i] == ubPage)		/* same page, then... */
			{
				return i;
			}
		}

		/* See if there is an unused physical page we can use. */

		for (i = 0; i < EMS_PAGES; i++)
		{
			if (fMappedBy[i] <= EMSMAPS)
			{
				MapPhysicalPage(i, uwHandle, ubPage);

			  	  uwPageEMMHandle[i] = uwHandle;
				ubPageLogicalPage[i] = ubPage;
						fMappedBy[i] = EMSMAPS;

				return i;
			}
		}

		/* Then use any old EMS page, as long as its not the same
			one that <pMem> is using. */

		for (i = 0; i < EMS_PAGES; i++)
		{
			if (!(FindPhysicalPage(pMem) <= i && i <= FindPhysicalPage(pMem+16383)))
			{
				MapPhysicalPage(i, uwHandle, ubPage);

				/* Assert: at this point, the variables uwPageEMMHandle[i],
					ubPageLogicalPage[i], and fMappedBy[i] do NOT contain
					the current logical to physical page mapping, rather it
					stores what the logical page should be paged back in
					after completing our read / write operation. */
				break;
			}
		}
		return i;
	}

} /* FindAndMapIntoPhysicalPage */


/*********************************************************************
 *
 * ActivateEMSatWin
 *
 * SYNOPSIS
 *		void far32 *ActivateEMSatWin (EMSPntr *ep, short window)
 *
 * PURPOSE
 *		Make object appear in page frame.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void far32 *ActivateEMSatWin (EMSPntr *ep, short window)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ActivateEMSatWin";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		union REGS	 inregs;
		union REGS	 outregs;
		short		 page;
		UWORD		 pages;
		UBYTE huge32	*data;

		pages = ep->Pages;
		for (page = 0; page < pages; page++)
		{
			fMappedBy[page+window]	= USERMAPS;

			/* If not already paged in, then page it in. */

			if (uwPageEMMHandle[page+window] != ep->EMMs[page].EMMHandle ||
				ubPageLogicalPage[page+window] != ep->EMMs[page].Page)
			{
				  uwPageEMMHandle[page+window]	= ep->EMMs[page].EMMHandle;
				ubPageLogicalPage[page+window]	= ep->EMMs[page].Page;

				inregs.h.ah = 0x44;
				inregs.h.al = page+window;
				inregs.x.bx = ep->EMMs[page].Page;
				inregs.x.dx = ep->EMMs[page].EMMHandle;
				int86 (0x67, &inregs, &outregs);
			}
		}
		LastEMS = ep;

		data = (UBYTE huge32 *)PageFramePtr +
				(UWORD) window * PAGE_SIZE +
				ep->EMMs[0].Offset;

		return (void far32 *)data;

	}

} /* ActivateEMSatWin */


/*********************************************************************
 *
 * SpecialFreeEMS
 *
 * SYNOPSIS
 *		void SpecialFreeEMS (EMSPntr *ep, short pages)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL void SpecialFreeEMS (EMSPntr *ep, short pages)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "SpecialFreeEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		short		 page;
		UWORD		 small;
		UWORD		 size;
		Book		*bk;
		FreeMemType	*fmt;
		FreeMemType	*nextfmt;
		EMMTracker	*emm;

		size = ep->Size;
		for (page = 0; page < pages; page++) {
			emm   = &ep->EMMs[page];
			small = min (size, PAGE_SIZE - emm->Offset);
			size -= small;
			bk    = Head (BookList);
			while (bk->EMMHandle != emm->EMMHandle) {
				bk = Next (bk);
			}
			fmt   = Head (bk->PageLists[emm->Page]);
			while (!EndOfList (fmt)) {
				if (emm->Offset + small < fmt->Offset) {
					nextfmt = CreateNode (sizeof (FreeMemType), NULL);
					if (!nextfmt) {
						SetGlobalErr (ERR_OUT_OF_MEMORY);
						GEcatf ("\n FreeEMS () no memory for FreeMemType");
						return;
					}
					nextfmt->Offset = emm->Offset;
					nextfmt->Size   = small;
					InsertBefore (fmt, nextfmt);
					break;
				} else if (emm->Offset + small == fmt->Offset) {
					fmt->Offset -= small;
					fmt->Size += small;
					nextfmt = Prev (fmt);
					if (!StartOfList (nextfmt)) {
						if (nextfmt->Offset + nextfmt->Size == fmt->Offset) {
							nextfmt->Size += fmt->Size;
							Remove (fmt);
							DeleteNode (fmt);
						}
					}
					break;
				} else if (emm->Offset == fmt->Offset + fmt->Size) {
					fmt->Size += small;
					nextfmt = Next (fmt);
					if (!EndOfList (nextfmt)) {
						if (fmt->Offset + fmt->Size == nextfmt->Offset) {
							fmt->Size += nextfmt->Size;
							Remove (nextfmt);
							DeleteNode (nextfmt);
						}
					}
					break;
				} else if (emm->Offset > fmt->Offset + fmt->Size) {
					if (EndOfList (Next(fmt))) {
						nextfmt = CreateNode (sizeof (FreeMemType), NULL);
						if (!nextfmt) {
							SetGlobalErr (ERR_OUT_OF_MEMORY);
							GEcatf ("\n FreeEMS () no memory for FreeMemType");
							return;
						}
						nextfmt->Offset = emm->Offset;
						nextfmt->Size   = small;
						InsertAfter (fmt, nextfmt);
						break;
					}
				}
				fmt = Next (fmt);
			}
		}

		FreeUnusedBooks ();

		DeleteEMSPntr (ep);
	}

} /* SpecialFreeEMS */


/*********************************************************************
 *
 * FreeEMS
 *
 * SYNOPSIS
 *		void FreeEMS (EMSPntr *ep)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void FreeEMS (EMSPntr *ep)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FreeEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	SpecialFreeEMS (ep, ep->Pages);

} /* FreeEMS */


/*********************************************************************
 *
 * AllocEMS
 *
 * SYNOPSIS
 *		EMSPntr *AllocEMS (size_t size)
 *
 * PURPOSE
 *		Allocate upto 64K of EMS memory;
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
EMSPntr *AllocEMS (size_t size)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "AllocEMS";
#endif
#if REQUIRE
	if (!size) {
		GEprintf ("Called AllocEMS() with a ZERO size");
		return NULL;
	}
#endif
/*-------------------------------------------------------------------------*/

	{
		short		 small;
		short		 page;
		short		 bpage;
		short		 found;
		short		 part;
		EMSPntr		*ep;
		Book		*bk;
		FreeMemType	*fmt;

 		if (!PageFrame) {
			if (!InitEMS ()) {
				return NULL;
			}
		}
		
		ep = NewEMSPntr();
		if (ep) {
			size = size + 15 & 0xFFF0;
			page = 0;
			ep->Size = size;
			if (size <= PAGE_SIZE) {
				small = TRUE;
			} else {
				small = FALSE;
			}
			while (size) {
				part  = min (size, PAGE_SIZE);
				bk    = Head (BookList);
				found = FALSE;
				while (!found && !EndOfList (bk)) {
					for (bpage = 0; bpage < bk->NumPages; bpage++) {
						fmt = Head(bk->PageLists[bpage]);
						while (!found && !EndOfList (fmt)) {
							if (small) {
								if (fmt->Size >= part) {
									found = TRUE;
									break;
								}
							} else {
								if (fmt->Size == PAGE_SIZE) {
									found = TRUE;
									break;
								}
							}
							fmt = Next (fmt);
						}
						if (found) {
							break;
						}
					}
					if (!found) {
						bk = Next (bk);
					}
				}
				if (!found) {
					if (!AddNewBook()) {
						goto aecleanup;
					}
					bk    = Tail (BookList);
					fmt   = Head (bk->PageLists[0]);
					bpage = 0;
				}
				ep->EMMs[page].EMMHandle = bk->EMMHandle;
				ep->EMMs[page].Page      = bpage;
				ep->EMMs[page].Offset    = fmt->Offset;
				ep->Pages++;
				fmt->Offset += part;
				fmt->Size   -= part;
				page++;
				size -= part;
			}
			return ep;
		}
aecleanup:
		if (ep)			SpecialFreeEMS (ep, page);
		return NULL;
	}

} /* AllocEMS */


/*********************************************************************
 *
 * ReadEMS
 *
 * SYNOPSIS
 *		void ReadEMS (EMSPntr *ep, UWORD uwOffset, UWORD uwLength,
 *					void far32 *pMem)
 *
 * PURPOSE
 *		Copies <uwLength> bytes from <uwOffset> from the start of
 *		EMS memory chunk <ep> to <pMem>.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void ReadEMS (EMSPntr *ep, UWORD uwOffset, UWORD uwLength, void far32 *pMem)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		short ip;
		UWORD uwMyOffset = 0;
		UWORD uwMyLength;

		for (ip = 0; ip < ep->Pages; ip++)
		{
			if (uwLength &&		/* If bytes remain to read, and... */
				(uwMyOffset <= uwOffset) &&	/* chunk falls within this page, then... */
				(uwOffset <= uwMyOffset+(PAGE_SIZE-1)))
			{
				UBYTE ubPage;			/* Physical EMS page used. */

				uwMyLength = min(uwMyOffset+PAGE_SIZE - uwOffset, uwLength);

				ubPage = FindAndMapIntoPhysicalPage(ep->EMMs[ip].EMMHandle, ep->EMMs[ip].Page, pMem);

				memcpy(pMem, PageFramePtr + (UWORD) ubPage * PAGE_SIZE + ep->EMMs[ip].Offset + (uwOffset-uwMyOffset), uwMyLength);

				if (fMappedBy[ubPage] == USERMAPS &&	/* If mapped by user, and... */
					 (ep->EMMs[ip].EMMHandle != uwPageEMMHandle[ubPage] || /* not same handle or */
					  ep->EMMs[ip].Page != ubPageLogicalPage[ubPage]))	/* not same logical page, then... */
				{
					/* Restore physical page to logical page assigned by user. */

					MapPhysicalPage(ubPage, uwPageEMMHandle[ubPage], ubPageLogicalPage[ubPage]);
				}

				(UBYTE *) pMem += uwMyLength;
				uwOffset += uwMyLength;
				uwLength -= uwMyLength;
			}
			uwMyOffset += PAGE_SIZE;
		}
	}

} /* ReadEMS */


/*********************************************************************
 *
 * WriteEMS
 *
 * SYNOPSIS
 *		void WriteEMS (void far32 *pMem, UWORD uwLength,
 *						EMSPntr *ep, UWORD uwOffset)
 *
 * PURPOSE
 *		Copies <uwLength> bytes from <pMem> to
 *		<uwOffset> from the start of EMS memory chunk <ep>.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void WriteEMS (void far32 *pMem, UWORD uwLength, EMSPntr *ep, UWORD uwOffset)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "WriteEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		short ip;
		UWORD uwMyOffset = 0;
		UWORD uwMyLength;

		for (ip = 0; ip < ep->Pages; ip++)
		{
			if (uwLength &&		/* If bytes remain to write, and... */
				(uwMyOffset <= uwOffset) &&	/* chunk falls within this page, then... */
				(uwOffset <= uwMyOffset+(PAGE_SIZE-1)))
			{
				UBYTE ubPage;			/* Physical EMS page used. */

				uwMyLength = min(uwMyOffset+PAGE_SIZE - uwOffset, uwLength);

				ubPage = FindAndMapIntoPhysicalPage(ep->EMMs[ip].EMMHandle, ep->EMMs[ip].Page, pMem);

				memcpy(PageFramePtr + (UWORD) ubPage * PAGE_SIZE + ep->EMMs[ip].Offset + (uwOffset-uwMyOffset), pMem, uwMyLength);

				if (fMappedBy[ubPage] == USERMAPS &&	/* If mapped by user, and... */
					 (ep->EMMs[ip].EMMHandle != uwPageEMMHandle[ubPage] || /* not same handle or */
					  ep->EMMs[ip].Page != ubPageLogicalPage[ubPage]))	/* not same logical page, then... */
				{
					/* Restore physical page to logical page assigned by user. */

					MapPhysicalPage(ubPage, uwPageEMMHandle[ubPage], ubPageLogicalPage[ubPage]);
				}

				(UBYTE *) pMem += uwMyLength;
				uwOffset += uwMyLength;
				uwLength -= uwMyLength;
			}
			uwMyOffset += PAGE_SIZE;
		}
	}

} /* WriteEMS */


/*********************************************************************
 *
 * AvailEMS
 *
 * PURPOSE
 *		Return available EMS memory in K bytes.
 *		If EMS memory has not been initialized, this routine
 *		returns -1, even if plenty of EMS exists.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Available EMS memory in K bytes, -1 if not used yet.
 *
 * HISTORY
 *		03/07/94 Monday (dcc) - add support for cEMSPagesAllocated
 *
 * SEE ALSO
 *
*/
long AvailEMS(void)
{

	long ulAvail = -1L;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "AvailEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (PageFrame)
	{
		Book *bk = Head (BookList);

		ulAvail = 0;
		while (!EndOfList (bk))
		{
			short bpage;

			for (bpage = 0; bpage < bk->NumPages; bpage++)
			{
				FreeMemType *fmt = Head (bk->PageLists[bpage]);

				while (!EndOfList (fmt))
				{
					ulAvail += fmt->Size;

					fmt = Next (fmt);
				}
			}
			bk = Next (bk);
		}
		ulAvail /= 1024;

		{
			union REGS	 regs;
			
			regs.h.ah = 0x42;
			int86 (0x67, &regs, &regs);

			/* See if there is a limit on how much EMS we may allocate */

			if (cEMSMaxPages)
				regs.x.bx = min(regs.x.bx, cEMSMaxPages - cEMSPagesAllocated);

			if (regs.h.ah == 0)
			{
				ulAvail += regs.x.bx * 16;
			}
		}
	}

	return ulAvail;
} /* AvailEMS */


/*********************************************************************
 *
 * PushEMS
 *
 * SYNOPSIS
 *		short PushEMS (void)
 *
 * PURPOSE
 *		Push the current EMS active pages
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short PushEMS (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "PushEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		EMSTracker		*ems;

		ems = CreateNode (sizeof (EMSTracker), NULL);
		if (!ems) {
			return FALSE;
		}
		ems->EMS = LastEMS;
		AddTail (EMSStack, ems);
		return TRUE;
	}

} /* PushEMS */

/*********************************************************************
 *
 * PopEMS
 *
 * SYNOPSIS
 *		void PopEMS (void)
 *
 * PURPOSE
 *		Pop the most resent Pushed ems pages
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void PopEMS (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "PopEMS";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		EMSTracker	*ems;

		ems = RemTail (EMSStack);
		ActivateEMS (ems->EMS);
		DeleteNode (ems);

	}

} /* PopEMS */

