/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * READINI.C
 *
 * PROGRAMMER : Dan Chang
 *    VERSION : 00.000
 *    CREATED : 01/19/92
 *   MODIFIED : 10/06/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		
 *
 * HISTORY
 *		01/19/92 (dcc) - modified from readCFG with more options
 *		09/18/92 (GAT) - added all the GetFirst and GetNext stuff
 *		09/18/92 (GAT) - put in MODULE stuff for library.
 *		02/27/94 (GAT) - Added CreateINI, AddINISection, AddINILine
 *		02/27/94 (GAT) - Changed most printfs to ErrMess
 *		03/02/94 (GAT) - Changed ReadINI to AppendINI and added
 *						 ReadINI macro
 *		10/06/94 (dcc) - updated to use <echidna/alloc.h>
 *	
*/

#include <echidna/platform.h>
#include "switches.h"

#include <string.h>
#include <stdio.h>
#include <echidna/alloc.h>
#include <echidna/strings.h>
#include <echidna/listfunc.h>
#include <echidna/readini.h>
#include <echidna/eerrors.h>
#include <echidna/eio.h>

/**************************** C O N S T A N T S ***************************/

#define MAX_LINE_SIZE	1024

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_readIni                                               */
/*------------------------------------------------------------------------*/

/****************************** G L O B A L S *****************************/

UWORD uwWarnErrCount;

static BOOL fCaseSensitive = TRUE;
static BOOL fSaveBlankLines = FALSE;
static BOOL fTrimWhiteSpace = TRUE;
static BOOL fStripComments = TRUE;
static BOOL fCommentAnywhere = TRUE;
static BOOL fMergeSections = TRUE;
static BOOL fExpandEVars = TRUE;
static char *szComment = ";";
static char *szSectionMarker = "[";

static UWORD uwLineCount;
static char szLine[MAX_LINE_SIZE];


/***************************** R O U T I N E S ****************************/

void SetINICaseSensitive(BOOL f)
{
	fCaseSensitive = f;
}

void SetINISaveBlankLines(BOOL f)
{
	fSaveBlankLines = f;
}

void SetINITrimWhiteSpace(BOOL f)
{
	fTrimWhiteSpace = f;
}

void SetINIStripComments(BOOL f)
{
	fStripComments = f;
}

void SetINICommentAnywhere(BOOL f)
{
	fCommentAnywhere = f;
}

void SetINIMergeSections(BOOL f)
{
	fMergeSections = f;
}

void SetINIComment(char *sz)
{
	szComment = sz;
}

void SetINISectionMarker(char *sz)
{
	szSectionMarker = sz;
}

UWORD GetINIWarnings(void)
{
	return uwWarnErrCount;
}

/*********************************************************************
 *
 * GetLine
 *
 * SYNOPSIS
 *		char *GetLine (char *line, size_t size, FILE *pFile)
 *
 * PURPOSE
 *		Read a line from a text file <pFile> into specifed buffer <line>.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static char *GetLine (char *line, size_t size, FILE *pFile)
{
	static char	 workLine[MAX_LINE_SIZE];
	char	*pch;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetLine";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	do
	{
		while ((pch = fgets (workLine, size, pFile)) != NULL)
		{
			size_t chCnt;

			uwLineCount++;

			if (fExpandEVars)
			{
				EIO_ExpandEVars (line, workLine, size);
			}
			else
			{
				strncpy (line, workLine, size);
				line[size - 1] = '\0';
			}

			if (fTrimWhiteSpace)
			{
				pch = TrimWhiteSpace(pch);
			}

			chCnt = strlen(pch);
			if (chCnt && pch[--chCnt] == '\n')
				pch[chCnt] = 0;

			if (fStripComments)
			{
				if (fCommentAnywhere)
				{
					char *pch2;

					if ((pch2 = strstr(pch, szComment)) != NULL)
					{
						*pch2 = 0;
						if (fTrimWhiteSpace)
						{
							pch = TrimWhiteSpace(pch);
						}
					}
				}
				else
				{
					if (strncmp(pch, szComment, strlen(szComment)) == 0)
						break;
				}
			}

			if (fSaveBlankLines || (!fSaveBlankLines && *pch))
				return pch;
		}
	} while (pch);
	return NULL;

} /* GetLine */

/*********************************************************************
 *
 * AddINILine
 *
 * SYNOPSIS
 *		ConfigLine * AddINILine (SectionTracker *pst, char *pszIniLine, UWORD lineNo)
 *
 * PURPOSE
 *		Add a new line to the specified section
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
ConfigLine *AddINILine (Section *pst, char *pszIniLine, UWORD lineNo)
{
	ConfigLine	*pLine;

	if ((pLine = CreateNode (sizeof (ConfigLine), pszIniLine)) != NULL)
	{
		pLine->LineNo = lineNo;
		AddTail(&pst->LineList, pLine);
	}

	return pLine;
} // AddINILine


/*********************************************************************
 *
 * AddINISection
 *
 * SYNOPSIS
 *		SectionTracker * AddINISection (IniList *pIniList, char *pszSectionName)
 *
 * PURPOSE
 *		Create a new empty section
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
Section *AddINISection (IniList *pIniList, char *pszSectionName)
{
	Section		*pSection;

	if ((pSection = CreateNode(sizeof (Section), pszSectionName)) != NULL)
	{
		AddTail(pIniList, pSection);

		InitList(&pSection->LineList);
	}

	return pSection;
} // AddINISection


/*********************************************************************
 *
 * CreateINI
 *
 * SYNOPSIS
 *		IniList * CreateINI (char *name)
 *
 * PURPOSE
 *		Create a new empty INI list
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
IniList *CreateINI (char *name)
{
	IniList	*pIniList;

	if ((pIniList = (IniList *) CreateList(name)) == NULL)
	{
		ErrMess ("OOM INI list.\n");
	}

	return pIniList;
} // CreateINI

/*********************************************************************
 *
 * AppendINI
 *
 * SYNOPSIS
 *		IniList * AppendINI (IniList *pIniList, char *filename)
 *
 * PURPOSE
 *		Append initialization file and store in appropriate linked lists.
 *
 * INPUT
 *		pIniList:			previous iniList or NULL
 *		filename:			File to read and store.
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Returns pointer to list of initialization sections,
 *		or NULL if unsuccessful.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
IniList *AppendINI(IniList *pIniList, char *filename)
{
	char		*pch;
	FILE		*pFile;

	ListType	*pCurrentList = NULL;
	Section		*pSection = NULL;

	uwLineCount = 0;
	uwWarnErrCount = 0;

	if ((pFile = fopen (filename, "r")) == NULL)
	{
		ErrMess ("Couldn't open configuration file '%s'\n", filename);
		return NULL;
	}

	if (!pIniList)
	{
		if ((pIniList = CreateINI(filename)) == NULL)
		{
			return NULL;
		}
	}

	/*** Read Config File ***/

	while ((pch = GetLine(szLine, MAX_LINE_SIZE, pFile)) != NULL)
	{
		if (strncmp (pch, szSectionMarker, strlen(szSectionMarker)) == 0)
		{
			pCurrentList = NULL;

			if (fMergeSections)
			{
				pSection = Head(pIniList);

				while (!IsEOList(pSection))
				{
					if ( (fCaseSensitive &&  strcmp(szSectionName(pSection), pch) == 0) ||
						(!fCaseSensitive && stricmp(szSectionName(pSection), pch) == 0))
					{
						pCurrentList = &pSection->LineList;
						break;
					}
					pSection = Next(pSection);
				}
			}
			if (!pCurrentList)
			{
				if ((pSection = AddINISection(pIniList, pch)) == NULL)
				{
					ErrMess ("OOM file '%s', line %d.\n", filename, uwLineCount);
/**/				goto ABORT;
				}

				pCurrentList = &pSection->LineList;
			}
		}
		else if (!pCurrentList)
		{
			WarnMess ("File %s, line %d: No section header.\n", filename, uwLineCount);
			uwWarnErrCount++;
		}
		else
		{
			if (AddINILine (pSection, pch, uwLineCount) == NULL)
			{
				ErrMess ("OOM file '%s', line %5d.\n", filename, uwLineCount);
/**/			goto ABORT;
			}
		}
	}

	fclose(pFile);
	return pIniList;

ABORT:
	FreeINI(pIniList);
	return NULL;


} // AppendINI

/*********************************************************************
 *
 * FindSection
 *
 * SYNOPSIS
 *		SectionTracker *FindSection(SectionTracker *pst,
 *										IniList *pIniList, char *szSectionName)
 *
 * PURPOSE
 *		Find 1st section that matches <szSectionName>, return
 *		pointer to list of configuration lines for that section.
 *
 * INPUT
 *		pst:					pointer to SectionTracker to fill in
 *		pIniList:			pointer to IniList of sections to search
 *		sz:					name of section to find
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		Returns pointer to a section tracker, or NULL if unsuccessful.
 *		Pass to GetNextINILine to get next line from section.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
SectionTracker *FindSection(SectionTracker *pst, IniList *pIniList, char *sz)
{
	if (pIniList)
	{
		Section *pSection;

		pSection = Head(pIniList);

		while (!IsEOList(pSection))
		{
			if ( (fCaseSensitive &&  strcmp(szSectionName(pSection), sz) == 0) ||
				(!fCaseSensitive && stricmp(szSectionName(pSection), sz) == 0))
			{
				pst->ltSection = &pSection->LineList;
				pst->clCurrentLine = (ConfigLine *) pst->ltSection;
				return pst;
			}
			pSection = Next(pSection);
		}
	}
	return NULL;
}


/*********************************************************************
 *
 * FindNextINILine
 *
 * SYNOPSIS
 *		char *FindNextINILine(SectionTracker *pst, char *sz)
 *
 * PURPOSE
 *		Find next config line from a section that STARTS WITH
 *		(matches) the string <sz>. Update variables to
 *		reflect current line being processed.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *
 * AFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		Returns pointer to 1st character in matching string
 *		following <sz>, or NULL if no more lines.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/

char *FindNextINILine(SectionTracker *pst, char *sz)
{
	ConfigLine *pLine;
	size_t uwLen;

	uwLen = strlen(sz);

	while ((pLine = GetNextINILine(pst)) != NULL)
	{
		if (( fCaseSensitive &&  strncmp(szConfigLine(pLine), sz, uwLen) == 0) ||
			(!fCaseSensitive && strnicmp(szConfigLine(pLine), sz, uwLen) == 0))
		{
			return szConfigLine(pLine)+uwLen;
		}
	}
	return NULL;
}

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_FreeINI                                               */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 *	FreeINI
 *
 * SYNOPSIS
 *		void FreeINI(IniList *pIniList)
 *
 * PURPOSE
 *		Free linked list structure generated by ReadINI.
 *
 * INPUT
 *		pIniList:			pointer to structure to free.
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void FreeINI(IniList *pIniList)
{
	if (pIniList)
	{
		Section *pSection;

		while ((pSection = RemTail(pIniList)) != NULL)
		{
			EmptyList(&pSection->LineList);

			DeleteNode(pSection);
		}
	}
	MEM_free(pIniList);
}

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_PrintINI                                              */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 *	PrintINI
 *
 * SYNOPSIS
 *		void PrintINI(IniList *pIniList)
 *
 * PURPOSE
 *		Print linked list structure generated by ReadINI.
 *
 * INPUT
 *		pIniList:			pointer to structure to print.
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void PrintINI(IniList *pIniList)
{
	if (pIniList)
	{
		Section *pSection;

		pSection = Head(pIniList);

		while (!IsEOList(pSection))
		{
			ConfigLine	*pLine;

			printf("'%s'\n", szSectionName(pSection));

			pLine = Head(&pSection->LineList);

			while (!IsEOList(pLine))
			{
				printf("  '%s'\n", szConfigLine(pLine));

				pLine = Next(pLine);
			}
			pSection = Next(pSection);
		}
	}
}

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_ResetSection                                          */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ResetSection
 *
 * SYNOPSIS
 *		void ResetSection(SectionTracker *pst)
 *
 * PURPOSE
 *		Resets <st> so GetNextINILine & FindNextINILine starts with
 *		the first config line.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to reset
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void ResetSection(SectionTracker *pst)
{
	pst->clCurrentLine = (ConfigLine *) pst->ltSection;
	return;
}

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetNextINILine                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetNextINILine
 *
 * SYNOPSIS
 *		ConfigLine *GetNextINILine(SectionTracker *pst)
 *
 * PURPOSE
 *		Get next config line from a section. Update variables to
 *		reflect current line being processed.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *
 * AFFECTS
 *		pst->clCurrentLine is updated to point to next line.
 *
 * RETURN VALUE
 *		Updates and returns pst->CurrentLine, which is a
 *		pointer to config line, or NULL if no more lines.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/

ConfigLine *GetNextINILine(SectionTracker *pst)
{
	if (pst->clCurrentLine == (ConfigLine *) pst->ltSection)
	{
		pst->clCurrentLine = Head(pst->ltSection);
	}
	else
	{
		pst->clCurrentLine = Next(pst->clCurrentLine);
	}
	if (!IsEOList(pst->clCurrentLine))
	{
		return pst->clCurrentLine;
	}
	else
	{
		return NULL;
	}
}

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_CountINILines                                         */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * CountINILines
 *
 * SYNOPSIS
 *		WORD CountINILines(SectionTracker *pst)
 *
 * PURPOSE
 *		Count number of config lines remaining in a section
 *		starting with current <pst>.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *
 * AFFECTS
 *
 * RETURN VALUE
 *		Returns number of config lines remaining in current section.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/

WORD CountINILines(SectionTracker *pst)
{
	ConfigLine *pLine;
	WORD wLines = 0;

	pLine = pst->clCurrentLine;			/* Save current line. */

	while(GetNextINILine(pst) != NULL)
		wLines++;

	pst->clCurrentLine = pLine;			/* Restore current line. */

	return wLines;
}

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetNextString                                         */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetNextString
 *
 * SYNOPSIS
 *		BOOL GetNextString (SectionTracker *pst, char *sz, char *dst, size_t maxlen)
 *
 * PURPOSE
 *		Find the next line from a section that STARTS WITH
 *		(matches) the string <sz> and copy at most <maxlen - 1>
 *		bytes from the line to the string pointed to by <dst>
 *		appending a '\0' to the end.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to string to fill in
 *		maxlen:				max number of bytes to affect in dst.
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetNextString (SectionTracker *pst, char *sz, char *dst, size_t maxlen)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetNextString";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		char	*s;

		s = FindNextINILine (pst, sz);
		if (s)
		{
			strncpy (dst, s, maxlen);
			dst[maxlen - 1] = '\0';
			return TRUE;
		}
		return FALSE;
	}

} /* GetNextString */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetNextShort                                          */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetNextShort
 *
 * SYNOPSIS
 *		BOOL GetNextShort (SectionTracker *pst, char *sz, short *dst)
 *
 * PURPOSE
 *		Find the next line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the short pointed to by
 *		<dst> to the value represented there.
 *
 *		If value starts with "0x" or "$" or ends with "h" the value
 *		will be converted as hex.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to short to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetNextShort (SectionTracker *pst, char *sz, short *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetNextShort";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		char	*s;

		s = FindNextINILine (pst, sz);
		if (s)
		{
			
			*dst = eggs_atos (s);
			return TRUE;
		}
		return FALSE;
	}

} /* GetNextShort */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetNextUShort                                         */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetNextUShort
 *
 * SYNOPSIS
 *		BOOL GetNextUShort (SectionTracker *pst, char *sz, USHORT *dst)
 *
 * PURPOSE
 *		Find the next line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the USHORT pointed to by
 *		<dst> to the value represented there.
 *
 *		If value starts with "0x" or "$" or ends with "h" the value
 *		will be converted as hex.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to USHORT to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetNextUShort (SectionTracker *pst, char *sz, USHORT *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetNextUShort";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		char	*s;

		s = FindNextINILine (pst, sz);
		if (s)
		{
			
			*dst = (USHORT)eggs_atol (s);
			return TRUE;
		}
		return FALSE;
	}
} /* GetNextUShort */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetNextLong                                           */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetNextLong
 *
 * SYNOPSIS
 *		BOOL GetNextLong (SectionTracker *pst, char *sz, long *dst)
 *
 * PURPOSE
 *		Find the next line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the long pointed to by
 *		<dst> to the value represented there.
 *
 *		If value starts with "0x" or "$" or ends with "h" the value
 *		will be converted as hex.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to long to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetNextLong (SectionTracker *pst, char *sz, long *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetNextLong";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		char	*s;

		s = FindNextINILine (pst, sz);
		if (s)
		{
			
			*dst = eggs_atol (s);
			return TRUE;
		}
		return FALSE;
	}
} /* GetNextLong */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetNextULong                                          */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetNextULong
 *
 * SYNOPSIS
 *		BOOL GetNextULong (SectionTracker *pst, char *sz, ULONG *dst)
 *
 * PURPOSE
 *		Find the next line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the ULONG pointed to by
 *		<dst> to the value represented there.
 *
 *		If value starts with "0x" or "$" or ends with "h" the value
 *		will be converted as hex.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to ULONG to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetNextULong (SectionTracker *pst, char *sz, ULONG *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetNextULong";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		char	*s;

		s = FindNextINILine (pst, sz);
		if (s)
		{
			
			*dst = eggs_atol (s);
			return TRUE;
		}
		return FALSE;
	}

} /* GetNextULong */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetNextBool                                           */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetNextBool
 *
 * SYNOPSIS
 *		BOOL GetNextULong (SectionTracker *pst, char *sz, BOOL *dst)
 *
 * PURPOSE
 *		Find the next line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the BOOL pointed to by
 *		<dst> to the value represented there.
 *
 *		Values for 0 are	: 0
 *							: FALSE
 *							: No
 *							: N
 *							: F
 *							: Off
 *
 *		Values for 1 are	: 1 - N
 *							: TRUE
 *							: Yes
 *							: T
 *							: Y
 *							: On
 *	
 *		Note: Checking for TRUE/FALSE/YES/NO is NOT case sensitive
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to BOOL to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
typedef struct {
	char	*s;
	BOOL	 val;
} BoolEntry;

static BoolEntry BoolTable[] = {
{	"0",		0,	},
{	"FALSE",	0,	},
{	"No",		0,	},
{	"N",		0,	},
{	"F",		0,	},
{	"Off",		0,	},
{	"1",		1,	},
{	"TRUE",		1,	},
{	"Yes",		1,	},
{	"T",		1,	},
{	"Y",		1,	},
{	"On",		1,	},
{	NULL,		0,  },
};

BOOL GetNextBool (SectionTracker *pst, char *sz, BOOL *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetNextBool";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		char	*s;

		s = FindNextINILine (pst, sz);
		if (s)
		{
			if (atoi (s))
			{
				*dst = TRUE;
				return TRUE;
			}

			{
				BoolEntry	*be;

				be = BoolTable;
				while (be->s)
				{
					if (!stricmp (be->s, s))
					{
						*dst = be->val;
						return TRUE;
					}
					be++;
				}
			}
		}
		return FALSE;
	}

} /* GetNextBool */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_FindFirstINILine                                      */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * FindFirstINILine
 *
 * SYNOPSIS
 *		char *FindFirstINILine (SectionTracker *pst, char *sz)
 *
 * PURPOSE
 *		Find the first line from a section that STARTS WITH
 *		(matches) the string <sz>.  Update variables to
 *		reflect current line being processed.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		Returns pointer to 1st character in matching string
 *		following <sz>, or NULL if no lines matches.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
char *FindFirstINILine (SectionTracker *pst, char *sz)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FindFirstINILine";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ResetSection (pst);
	return FindNextINILine (pst, sz);

} /* FindFirstINILine */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetFirstString                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetFirstString
 *
 * SYNOPSIS
 *		BOOL GetFirstString (SectionTracker *pst, char *sz, char *dst, size_t maxlen)
 *
 * PURPOSE
 *		Find the first line from a section that STARTS WITH
 *		(matches) the string <sz> and copy at most <maxlen - 1>
 *		bytes from the line to the string pointed to by <dst>
 *		appending a '\0' to the end.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to string to fill in
 *		maxlen:				max number of bytes to affect in dst.
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetFirstString (SectionTracker *pst, char *sz, char *dst, size_t maxlen)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetFirstString";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ResetSection (pst);
	return GetNextString (pst, sz, dst, maxlen);

} /* GetFirstString */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetFirstShort                                         */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetFirstShort
 *
 * SYNOPSIS
 *		BOOL GetFirstShort (SectionTracker *pst, char *sz, short *dst)
 *
 * PURPOSE
 *		Find the first line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the short pointed to by
 *		<dst> to the value represented there.
 *
 *		If value starts with "0x" or "$" or ends with "h" the value
 *		will be converted as hex.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to short to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetFirstShort (SectionTracker *pst, char *sz, short *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetFirstShort";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ResetSection (pst);
	return GetNextShort (pst, sz, dst);

} /* GetFirstShort */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetFirstUShort                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetFirstUShort
 *
 * SYNOPSIS
 *		BOOL GetFirstUShort (SectionTracker *pst, char *sz, USHORT *dst)
 *
 * PURPOSE
 *		Find the first line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the USHORT pointed to by
 *		<dst> to the value represented there.
 *
 *		If value starts with "0x" or "$" or ends with "h" the value
 *		will be converted as hex.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to USHORT to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetFirstUShort (SectionTracker *pst, char *sz, USHORT *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetFirstUShort";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ResetSection (pst);
	return GetNextUShort (pst, sz, dst);

} /* GetFirstUShort */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetFirstLong                                          */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetFirstLong
 *
 * SYNOPSIS
 *		BOOL GetFirstLong (SectionTracker *pst, char *sz, long *dst)
 *
 * PURPOSE
 *		Find the first line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the long pointed to by
 *		<dst> to the value represented there.
 *
 *		If value starts with "0x" or "$" or ends with "h" the value
 *		will be converted as hex.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to long to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetFirstLong (SectionTracker *pst, char *sz, long *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetFirstLong";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ResetSection (pst);
	return GetNextLong (pst, sz, dst);

} /* GetFirstLong */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetFirstULong                                         */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetFirstULong
 *
 * SYNOPSIS
 *		BOOL GetFirstULong (SectionTracker *pst, char *sz, ULONG *dst)
 *
 * PURPOSE
 *		Find the first line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the ULONG pointed to by
 *		<dst> to the value represented there.
 *
 *		If value starts with "0x" or "$" or ends with "h" the value
 *		will be converted as hex.
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to ULONG to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetFirstULong (SectionTracker *pst, char *sz, ULONG *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetFirstULong";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ResetSection (pst);
	return GetNextULong (pst, sz, dst);

} /* GetFirstULong */

/*------------------------------------------------------------------------*/
/**# MODULE:READINI_GetFirstBool                                          */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * GetFirstBool
 *
 * SYNOPSIS
 *		BOOL GetFirstULong (SectionTracker *pst, char *sz, BOOL *dst)
 *
 * PURPOSE
 *		Find the first line from a section that STARTS WITH
 *		(matches) the string <sz> and sets the BOOL pointed to by
 *		<dst> to the value represented there.
 *
 *		Values for 0 are	: 0
 *							: FALSE
 *							: No
 *							: N
 *							: F
 *
 *		Values for 1 are	: 1 - N
 *							: TRUE
 *							: Yes
 *							: T
 *							: Y
 *	
 *		Note: Checking for TRUE/FALSE/YES/NO is NOT case sensitive
 *
 * INPUT
 *		pst:				pointer to SectionTracker to get lines from
 *		sz:					name of configuration to match
 *		dst:				pointer to BOOL to fill in
 *
 * EFFECTS
 *		pst->clCurrentLine is updated to point to line that starts with <sz>.
 *
 * RETURN VALUE
 *		returns TRUE of dst was affected.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL GetFirstBool (SectionTracker *pst, char *sz, BOOL *dst)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "GetFirstBool";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ResetSection (pst);
	return GetNextBool (pst, sz, dst);

} /* GetFirstBool */

