/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * BRUSHSUP.c
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 12/20/89
 *   MODIFIED : 10/07/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Support routines for BRUSHPAK files.
 *
 * HISTORY
 *		05/29/91 (RGM) - Changes for Amiga library under
 *						 SAS/Lattice C 5.10a.
 *		10/07/94 (dcc) - updated to use <echidna/alloc.h>
 *
*/
#include <echidna/platform.h>
#include "switches.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <echidna/etypes.h>
#include <echidna/alloc.h>
#include <echidna/eerrors.h>
#include <echidna/memory.h>
#include <echidna/graphics.h>
#include <echidna/eio.h>
#include <echidna/brushsup.h>
#include <echidna/fpfload.h>
#include <echidna/stdload.h>
#include <echidna/pixlscan.h>
#include <echidna/bpcfg.h>
#include "xpaksup.h"
#include "bpmasks.h"

/**************************** C O N S T A N T S ***************************/
#define NUM_CGA_COLORS	4
#define MAX_PACKED_SHAPE_SIZE	64000L

/***************************** T Y P E D E F S ****************************/
typedef struct {
	UWORD 	 NumShapes;
	UWORD 	 NumMCGAColors;
	UWORD 	 NumEGAColors;
	UWORD 	 CGAPaletteCode;
	UWORD 	 NumRemapTables;
	UWORD 	 NumEGAPatmapTables;
	UWORD 	 NumCGAPatmapTables;
	UWORD	 GlobalEGAPatmapIndex;
	UWORD	 GlobalCGAPatmapIndex;
} BPIHeaderType;

typedef struct {
	RemapInfoType		 RemapInfo;
#if __MSDOS__
	PatternInfoType		 PatternInfo;
#endif
} BPIInfoType;	
	
/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/
#if __iAPx86__
#define makeid(d,c,b,a)   (((LONG)(a)<<24)|((LONG)(b)<<16)|((c)<<8)|(d))
#else
#define makeid(a,b,c,d)   (((LONG)(a)<<24)|((LONG)(b)<<16)|((c)<<8)|(d))
#endif

#define CHUNK_HEAD	makeid ('H','E','A','D')
#define CHUNK_PBOX	makeid ('P','B','O','X')
#define CHUNK_MPAL	makeid ('M','P','A','L')
#define CHUNK_EPAL	makeid ('E','P','A','L')
#define CHUNK_RMAP	makeid ('R','M','A','P')
#define CHUNK_EPRM	makeid ('E','P','R','M')
#define CHUNK_CPRM	makeid ('C','P','R','M')
#define CHUNK_BATT	makeid ('B','A','T','T')
#define CHUNK_BINF	makeid ('B','I','N','F')
#define CHUNK_BDAT	makeid ('B','D','A','T')


/***************************** R O U T I N E S ****************************/

/*********************************************************************
 *
 * LoadnTranslate
 *
 * SYNOPSIS
 *		short	LoadnTranslate (
 *			BPIType		*bpi,
 *			BPIInfoType	*bpi_info,
 *			UWORD		 numshapes
 *		)
 *
 * PURPOSE
 *		Load and tranform and translate shapes.
 *		Transforming involves flipping remapping and duplicating.
 *      There are 2 kinds of duplication, Copying and Kloning.
 *      To understand Copying and Kloning, you must first understand
 *      the 3 stages a of the coming into being of a Shape at Game Load Time.
 *  
 *          stage 1: Raw data that was saved by Brushpak in a '.BPI' file is 
 *                  loaded into a static buffer, B.
 *          stage 2: Raw data from B is copied to another static buffer, B2,
 *                  but in the process is transformed (i.e. remapped, 
 *                  and or flipped about horizontally  and or vertically).
 *          stage 3: The transformed data is translated into the graphics
 *                  format appropriate to the machine (MCGA, EGA or CGA) and
 *                  is now a Shape (structure useable by Echidna graphics
 *                  routines). This data is stored in an allocated buffer,
 *                  B3 (B and B2 are not modified).
 *
 *      Obfuscating Note:
 *          These stages are almost but not quiet completely totally 
 *          unlike the three stages in the metamorphic life cycle of a
 *          frog (egg, tadpole, adult).
 *      
 *      Definition of Copy:
 *          Let A and B be brushes.
 *			IF Stage 3 of A can be created from stage 1 of B by a 
 *              combination of remapping and flipping,
 *          THEN  A is a 'Copy' of B AND 
 *				A will be marked as a 'Copy' and B will be marked 
 *				as 'Copied' in their Attributes.
 *	
 *      Definintion of Klone:
 *          Let A and B be brushes.
 *          IF Stage 3 of A is identical to stage 3 of B,
 *          THEN A is a 'Klone' of B AND
 *				A will be marked as a 'Klone' and B will be marked 
 *				as 'Kloned' in their Attributes.
 *
 *	
 * Note: Brushpak ensures that all Copies come from source brushes that are
 *	not themselves Copies.  So the decsendants of a source brush presented 
 *	in tree form will always have height 1.
 * example:
 *
 *       source
 *       |||||
 *  +----+|||+----+
 *  |  +--+|+--+  |
 *  |  |   |   |  |
 *  c1 c2  c3  c4 c5
 *
 * Example of tree of height > 1 wich brushpak will never generate:
 *       source
 *        |  
 *        c1 
 *        |  
 *        c2 
 *
 * Note2: Brushpak ensures that all Klones come from source brushes that are
 *	not themselves Klones but may be Copies.
 *
 *
 * Note3: Being a Copy and being a Klone are mutally exclusive.
 * Here are the Possible combos of the four flags in any brush 
 * (X=TRUE, .=FALSE):
 *    Copy  Klone  Copied  Kloned
 *    ----- -----  ------  ------
 *      X    .      .        .     Copy  
 *    ---------------------------
 *      .    X      .        .     Klone
 *    ---------------------------
 *      .    .      X        .     Copied
 *    ---------------------------
 *      .    .      .        X     Kloned
 *    ---------------------------
 *      X    .      .        X     Copy|Kloned
 *    ---------------------------
 *      .    .      X        X     Copied|Kloned
 *    ---------------------------
 * HISTORY
 *	03/07/90 Wednesday (JMA) - Added Null bit attribute setting of processed
 *				brushes to keep them from being reprocessed my mistake.
 *				I will use the facts that being a copy and being copied are
 *				mutually exclusive and that being a klone and being kloned
 *				are mutually exclusive to mark a brush as having already
 *				been processed.  So:
 *				Copy  Klone    Meaning
 *				----  -----	   -------
 *				  X     .      Is a copy that has NOT been processed.
 *				  X     X      Is a copy that has been processed.
 *				  .     X      Is a klone that has NOT been processed.
 *				  X     X      Is a klone that has been processed.    
 *
 *	04/24/90 Tuesday (JMA) - Added bpi_info parameter. It has the remap
 *				and pattern data pointers in it instead of having them
 *				in the bpi sturcture since they should be freed after
 *				this routine is called.
 *
*/
short	LoadnTranslate (
 	BPIType		*bpi,
	BPIInfoType	*bpi_info,
 	UWORD		 numshapes
)
{
	UWORD	 i, k, c;
	ULONG	*attribptr = bpi->AttribTable;
	ULONG	*kattribptr, *cattribptr;
	ULONG	 attrib, kattrib, cattrib;
	ULONG	*infoptr = (ULONG *)bpi->Shapes;
	ULONG	*kinfoptr, *cinfoptr;
	ULONG	 info, kinfo, cinfo;	/* remap index and size of shape */
	GShape	*newshapeptr;
	RemapTableType	
			*remap = NULL;
	short		result = TRUE;
	ConvertSType	cs;
#if __MSDOS__
	short	patindex, prevpatindex = -1;
#endif

	if (! OpenPixlScan (&cs, MAX_COLORS)) {
		return FALSE;
	}

	for (i = 0; i < numshapes; i++) {
		attrib	= *attribptr++; 
		info	= *infoptr++;		
		if (IsNull (attrib)) {
			bpi->Shapes[i] = NULL;
			continue;
		}
		if (! IsKlone (attrib)) { 
            if (! IsCopy (attrib)) { /* Is original, process it */

				/* Read shape into Raw buffer, Stage 1 */
				{
					int count;
                	if ((count = (int) FPF_read (RawBuf, (size_t)ShapeSize (info)))
						== -1 || (unsigned)count  != ShapeSize (info))
					{
						GEcatf ("\nLoadnTranslate:reading shape");
						result = FALSE;
						goto DONE;
					}
				}
                /* Set Up Remap Table if necessary, Stage 2 */
				if (IsRemapped (attrib)) {
					remap = &bpi_info->RemapInfo.Tables[(size_t)RemapIndex(info)];
					SetColorMap (remap, cs.ColorMap);
					SetRemapFlag (cs,TRUE);
				} else {
					SetRemapFlag (cs,FALSE);
				}

				/** Translate Brush, Stage 2 and 3 **/
				switch ((short)UserCode (attrib)) {
				case DONT_TOUCH:
					newshapeptr = GetCopy (RawBuf, (size_t)ShapeSize(info));	
					break;
				case MAKE_GSHAPE:
#if __MSDOS__
					switch (GraphicsMode) {
					case MODE_EGA: case MODE_TANDY:
						cs.PatternMap = &bpi_info->PatternInfo.PatmapTables[
							(patindex = (short)EGAPatmapIndex(attrib))];
						break;
					case MODE_CGA:
						cs.PatternMap = &bpi_info->PatternInfo.PatmapTables[(short)
							CGAPatmapIndex(attrib)];
						break;
					}	

					/* Translate */
					cs.ExpandPatterns = (patindex != prevpatindex);
					prevpatindex = patindex;
					cs.pcc = (UBYTE)PCC (attrib);
#endif
					cs.ShapeData = (GShape *)RawBuf;
					cs.HFlip = (short)IsHflipped (attrib);
					cs.VFlip = (short)IsVflipped (attrib);
					if ( (newshapeptr = ConvertGShape (&cs)) == NULL)
					{
						result = FALSE;
						goto DONE;
					}
					break;
				default:
					SetGlobalErr (ERR_GENERIC);
					GEprintf1 ("LoadnTranslate: Bad shape type %d", (short)UserCode (attrib));
					result = FALSE;
					goto DONE;
				}
				if (IsRemapped (attrib)) {
					ClearColorMap (remap, cs.ColorMap);
				} 
				bpi->Shapes[i] = newshapeptr;
			}

			/** Make Klones **/
            if (IsKloned (attrib)) { /* i May be Copy or Original */
    			k = i + 1;
				kattribptr	= attribptr;
				kinfoptr	= infoptr;
    			while (k < numshapes) {
					kattrib	= *kattribptr++;
					kinfo	= *kinfoptr++;
        			if (IsKlone (kattrib) && ! IsProcessedKlone (kattrib) &&
						KloneSource (kinfo) == i) 
					{
            			bpi->Shapes[k] = newshapeptr;
						kattribptr[-1] |= ATT_PROCESSED_KLONE_FLAG;
        			}
        			++k;
    			}
            }

			/** Make Copies **/
            if (IsCopied (attrib)) { /* Find and process Copies */
    			c = i + 1;
				cattribptr	= attribptr;
				cinfoptr	= infoptr;
    			while (c < numshapes) {
					cattrib	= *cattribptr++;
					cinfo	= *cinfoptr++; 
        			if (IsCopy (cattrib) && ! IsProcessedCopy (cattrib) &&
						CopySource (cinfo) == i) 
					{
                		/* Set up Remap Required, Stage 2 */
						if (IsRemapped (cattrib)) {
							remap = &bpi_info->RemapInfo.Tables[
								(size_t)RemapIndex(cinfo)];
							SetColorMap (remap, cs.ColorMap);
							SetRemapFlag (cs,TRUE);
						} else {
							SetRemapFlag (cs,FALSE);
						}

						/** Translate Brush Copy, Stage 2 and 3 **/
						switch ((short)UserCode (cattrib)) {
						case DONT_TOUCH:
							newshapeptr = GetCopy (RawBuf, 
								(size_t)ShapeSize(cinfo));	
							break;
						case MAKE_GSHAPE:
#if __MSDOS__
							switch (GraphicsMode) {
							case MODE_EGA: case MODE_TANDY:

								cs.PatternMap = &bpi_info->PatternInfo.
									PatmapTables[(patindex = 
										(short)EGAPatmapIndex(cattrib))];
								break;
							case MODE_CGA:
								cs.PatternMap = &bpi_info->PatternInfo.
									PatmapTables[(short)
									CGAPatmapIndex(cattrib)];
								break;
							}
	
							/* Translate */
							cs.ExpandPatterns = (patindex != prevpatindex);
							prevpatindex = patindex;
							cs.pcc = (UBYTE)PCC (cattrib);
#endif
							cs.ShapeData = (GShape *)RawBuf;
							cs.HFlip = (short)IsHflipped (cattrib);
							cs.VFlip = (short)IsVflipped (cattrib);
							if ( (newshapeptr = ConvertGShape (&cs)) == NULL)
							{
								result = FALSE;
								goto DONE;
							}
							break;
						default:
							SetGlobalErr (ERR_GENERIC);
							GEprintf1 ("LoadnTranslate: Bad shape type %d", (short)UserCode (attrib));
							result = FALSE;
							goto DONE;
						}
						if (IsRemapped (cattrib)) {
							ClearColorMap (remap, cs.ColorMap);
						} 
						bpi->Shapes[c] = newshapeptr;
						cattribptr[-1] |= ATT_PROCESSED_COPY_FLAG;
        			}
					++c;
    			}
            }
        }   /* Is a Klone and would already have been taken care of when the
            * source of the Klone was processed.
            */
		++bpi->NumShapes;
    }

DONE:
	ConvertGShape (NULL); /* So it can shut down (free buffers etc.) */
	ClosePixlScan (&cs, MAX_COLORS);
	return result;
	
} /* LoadnTranslate */

/*********************************************************************
 *
 * LoadBPI
 *
 * SYNOPSIS
 *		BPIType	*LoadBPI (
 *			char		*filename,
 *			short			 fpfindex
 *		)
 *
 * USAGE
 *		BPIptr = LoadBPI ("LostMaze.BPI");
 *		if (!BPIptr)
 *			printf ("LoadBPI Failed; ErrorCode = %d\n", GlobalErr);
 *
 * PURPOSE
 *		To load all the image data from a BPI (Brush Pack Images) file
 *		and convert the images data upon loading into the appropriate graphic
 *		mode format.  To allocate and fill in a BPIType structure and fill
 *		it in with the shape table and palette table (if present), and other
 *		information for use when freeing the data.  To free the data, pass 
 *		the pointer returned by this routine to FreeBPI().
 *
 * INPUTS
 *		filename:	Name of BrushPak created BPI (Brush Pack Images) 
 *					file.
 * RESULTS
 *		Returns pointer to BPIType structure if succesful. Else returns 
 *		NULL and sets GlobalErr to error code.
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BPIType	*LoadBPI (
	char		*filename,
	short			 fpfindex
)
{
	short				 i;
	short				 numtables;
	UWORD				 numshapes = 0;
	BPIType				*bpi;
	BPIHeaderType		header;
	ULONG				*sizeinfo;
	ChunkType			 chunk;
	size_t				 size1, size2, size3, size4;
	RemapTableType		*remaptableptr;
	UWORD				*uwordptr;
	UBYTE				*ubyteptr;
#if __MSDOS__
	short				 p;
	size_t				 sizetablesize, pattabarraysize = 0, totalpatdatsize;
	size_t				 spacetoalloc = 0;
	UWORD				 numpats;
	PatmapTableType		*patmaptableptr;
	UWORD				*tablesizes;
	UBYTE				*patdatptr, *patdatloadptr;
	UBYTE				*oddpatptr, *evenpatptr;
	UBYTE				 cgapat;
	UBYTE				 temp;
#endif
	size_t				 remapbuf1 = 0;
	size_t				 remapbuf2 = 0;
	UBYTE				 foundMPAL = FALSE;
	short				 GotBDAT   = FALSE;
	BPIInfoType			 bpi_info;

	/** Allocate BPIType tracking structure **/
	if ((bpi = (BPIType *)MEM_calloc (1, sizeof (BPIType))) == NULL)
	{
		SetGlobalErr ( ERR_OUT_OF_MEMORY);
		GEprintf2 ("LoadBPI:OOM allocating BPIType:'%s'-%d", filename, fpfindex);
		return NULL;
		
	}

	/** Open File **/
	if (! FPF_open (filename, fpfindex))
	{
		/* report error msg from FPF_open */
		goto FREE_EXIT;
	}


	/** Read And Process Chunks Until Final Chunk (BDAT) is Read **/
	while (! GotBDAT) {
		/** Read Chunk Label and Size **/
		if ( (FPF_read (&chunk, sizeof (ChunkType))) !=
			sizeof (ChunkType)) 
		{
			GEcatf ("\nLoadBPI:Error reading CHUNK header:");
			goto CLOSE_EXIT;
		}

		/** Identify and Process Chunk */
		if (chunk.ChunkLabel == CHUNK_HEAD) {		 /** H E A D **/	
			if ( (FPF_read (&header, sizeof (BPIHeaderType))) != 
				sizeof (BPIHeaderType))
			{
				GEcatf ("\nLoadBPI:Error reading HEAD Chunk:");
				bpi->NumShapes = 0;
				goto CLOSE_EXIT;
			}
			numshapes = header.NumShapes;

#if __MSDOS__
			if (GraphicsMode == MODE_CGA) {
				bpi->Palette.NumColors = NUM_CGA_COLORS;
				bpi->Palette.Data.CGA = header.CGAPaletteCode;
			}
#endif
		} else if (chunk.ChunkLabel == CHUNK_PBOX)  /** P B O X **/
		{
			if ((bpi->PaintBox = ReadPBX ((size_t)chunk.ChunkSize)) == NULL)
			{
				GEcatf ("\nLoadBPI:reading CCSI chunk");
				goto CLOSE_EXIT;
			}
		}
		 else if (chunk.ChunkLabel == CHUNK_MPAL) { /** M P A L **/
			foundMPAL = TRUE;
#if __MSDOS__
			if (GraphicsMode == MODE_MCGA)
#endif
			{
				/* Allocate RGB Palette */
				if ((bpi->Palette.Data.RGB = (ColorType *)MEM_malloc (
					(size_t)chunk.ChunkSize)) == NULL)
				{
					SetGlobalErr ( ERR_OUT_OF_MEMORY);
					GEprintf2 ("LoadBPI:OOM allocating Palette table:MPAL(1) '%s'-%d", filename, fpfindex);
					goto CLOSE_EXIT;
				}
				/* Read Palette */
				if ( (FPF_read (bpi->Palette.Data.RGB, 
					(size_t)chunk.ChunkSize)) !=	chunk.ChunkSize) 	
				{
					MEM_free (bpi->Palette.Data.RGB);
					GEcatf ("\nLoadBPI:Error reading MPAL(2) Chunk");
					goto CLOSE_EXIT;
				}
				bpi->Palette.NumColors = header.NumMCGAColors;
			}
#if __MSDOS__
			 else {
				if (! FPF_skip (chunk.ChunkSize)) {
					SetGlobalErr ( ERR_READING_FILE);
					GEprintf2 ("LoadBPI:Error skipping MPAL/size:'%s'-%d", filename, fpfindex);
					goto CLOSE_EXIT;
				}
			}
#endif
		}
#if __MSDOS__
		 else if (chunk.ChunkLabel == CHUNK_EPAL &&
					(GraphicsMode == MODE_EGA || 
					 GraphicsMode == MODE_TANDY))	{ /** E P A L **/
			/* Allocate EGA Palette */
			if ((bpi->Palette.Data.EGA = MEM_malloc ((size_t)chunk.ChunkSize)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadBPI:OOM allocating EGA palette:EPAL(1) '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
			/* Read Palette */
			if ( (FPF_read (bpi->Palette.Data.EGA, 
				(size_t)chunk.ChunkSize)) != chunk.ChunkSize) 
			{
				MEM_free (bpi->Palette.Data.EGA);
				GEcatf ("\nLoadBPI:Error reading EPAL(2) Chunk");
				goto CLOSE_EXIT;
			}
			bpi->Palette.NumColors = header.NumEGAColors;
		}
#endif
		 else if (chunk.ChunkLabel == CHUNK_RMAP)	{ /** R M A P **/
			numtables
			= bpi_info.RemapInfo.NumRemapTables
			= header.NumRemapTables;
			size1 = remapbuf1 = numtables * sizeof (RemapTableType);

			/* Allocate Buffer for Remap Table Structures */
			if ((bpi_info.RemapInfo.Tables = (RemapTableType *)
				CreateTempBuffer (size1)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadBPI:OOM allocating RemapTables:RMAP(1) '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
			remaptableptr = bpi_info.RemapInfo.Tables;

			uwordptr			= (UWORD *)remaptableptr;
			size2				= numtables * sizeof (UWORD);
#if LATTICE
			{
				UBYTE	*ubp_temp	= (UBYTE *) uwordptr;
				ubp_temp			+= (size1 - size2);
				uwordptr			= (UWORD *) ubp_temp;
			}
#else
			uwordptr				= (UWORD *) ((UBYTE *) uwordptr +
									   (size1 - size2));
#endif

			/** Read Remap Table Sizes into Low End of Buffer **/
			if ( (FPF_read (uwordptr, size2) != size2)) 
			{
				GEcatf ("\nLoadBPI:Error reading RMAP(2) Chunk");
				goto CLOSE_EXIT;
			}
			
			/** Get Total size of Remap Tables **/
			size3 = remapbuf2 = (size_t)chunk.ChunkSize - size2;

			/** Allocate Buffer2 for All Remap Table Data **/
			if ((ubyteptr = (UBYTE *)CreateTempBuffer (size3)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadBPI:OOM allocating RemapTables:RMAP(3) '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}

			/** Read All Remap Tables into Buffer **/
			if ((FPF_read (ubyteptr, size3)) != size3) 
			{
				MEM_free (ubyteptr);
				GEcatf ("\nLoadBPI:Error reading RMAP(4) Chunk");
				goto CLOSE_EXIT;
			}
			
			/** Set Pointers in RemapTable Struct to Data in Buffer2 **/
			for (i = 0; i < numtables; i++) {
				remaptableptr->Elements = size4 = (*uwordptr) / 2;
				remaptableptr->From = ubyteptr;
				ubyteptr += size4;
				remaptableptr->To = ubyteptr;
				ubyteptr += size4;
				++uwordptr;
				++remaptableptr;
			}
		}
#if __MSDOS__
		 else if ((chunk.ChunkLabel == CHUNK_EPRM &&
						(GraphicsMode == MODE_EGA || 
					 		GraphicsMode == MODE_TANDY)) ||
					(chunk.ChunkLabel == CHUNK_CPRM &&
						GraphicsMode == MODE_CGA))	{	/** E P R M **/
														/** C P R M **/
			numtables 
			= bpi_info.PatternInfo.NumPatmapTables 
			= (chunk.ChunkLabel == CHUNK_EPRM)? header.NumEGAPatmapTables :
				header.NumCGAPatmapTables;
			pattabarraysize = numtables * sizeof (PatmapTableType);

			/* Allocate Buffer for Patmap Table Structures */
			if ((patmaptableptr = (PatmapTableType *)CreateTempBuffer 
				(pattabarraysize)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadBPI:OOM allocating patterntables:?PRM(1) '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
			bpi_info.PatternInfo.PatmapTables = patmaptableptr;

			/** Read Patmap Table Sizes into Low End of Buffer **/
			/** Only works if sizeof(PatmapTableType) >= 2 * sizeof(UWORD) */
			sizetablesize		= numtables * sizeof (UWORD);
			tablesizes			= (UWORD *) patmaptableptr;
			tablesizes			= (UWORD *) ((UBYTE *) tablesizes +
								   (pattabarraysize - sizetablesize));

			if ( (FPF_read (tablesizes, sizetablesize) != sizetablesize)) 
			{
				GEcatf ("\nLoadBPI:Error reading ?PRM(2) Chunk");
				goto CLOSE_EXIT;
			}

			/** Get Total size of Patmap Tables **/
			spacetoalloc = 
				totalpatdatsize = 
					(size_t)chunk.ChunkSize - sizetablesize;

			/* Double Allocated space for CGA for unpacking bytes */
			if (chunk.ChunkLabel == CHUNK_CPRM ) {
				spacetoalloc *= 2;
			}

			/** Allocate Buffer2 for All EGA Patmap Table Data **/
			if ((patdatptr = (UBYTE *)CreateTempBuffer (spacetoalloc)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadBPI:OOM allocating pattern tables:?PRM(3) '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
			bpi_info.PatternInfo.DataBuf = patdatloadptr = patdatptr;

			/** For CGA, read patterns into lower half of buffer **/
			if (chunk.ChunkLabel == CHUNK_CPRM ) {
				patdatloadptr += totalpatdatsize;
			} 

			/** Read All Patmap Tables into Buffer **/
			if ((FPF_read (patdatloadptr, totalpatdatsize)) != 
				totalpatdatsize) 
			{
				MEM_free (patdatptr);
				GEcatf ("\nLoadBPI:Error reading ?PRM(4) Chunk");
				goto CLOSE_EXIT;
			}


			/** Set Pointers in PatmapTable Structs to Data in Buffer2 **/
			for (i = numtables; i; i--) {
				numpats = *tablesizes++;

				/* Unpack and move CGA pattern arrays */
				if (chunk.ChunkLabel == CHUNK_CPRM) {

					/* Get src and dest ptrs for unpacking CGA patterns */
					oddpatptr = patdatptr;
					evenpatptr = patdatptr + numpats;

					for (p = numpats; p ; p--) {
						cgapat = *patdatloadptr++; /* Get packed pat byte */

						/* Copy odd line pat from low nybble */
						temp = cgapat & 0x0F;
						*oddpatptr++ = temp | (temp << 4);

						/* Copy even line pat from high nybble */
						temp = cgapat & 0xF0;
						*evenpatptr++ = temp | (temp >> 4);
					}
				} else { /* EGA pats are a word */
					numpats /= 2; /* size of table / size of pattern */
				}
				patmaptableptr->NumPatterns = numpats;
				patmaptableptr->Odd = patdatptr;
				patdatptr += numpats;
				patmaptableptr->Even = patdatptr;
				patdatptr += numpats;
				++patmaptableptr;
			}

			/* Fill in Global Pattern Table */
			switch (GraphicsMode) {
			case MODE_EGA: case MODE_TANDY:
				i = header.GlobalEGAPatmapIndex;
				break;
			case MODE_CGA:
				i = header.GlobalCGAPatmapIndex;
				break;
			}
			if (i--) {
				patmaptableptr = &bpi_info.PatternInfo.PatmapTables[i];
				numpats = patmaptableptr->NumPatterns;
				SetGlobalPattern (patmaptableptr);

				/* Get copy for bpi structure */
				bpi->GlobalPattern.NumPatterns = numpats;
				bpi->GlobalPattern.Even = GetCopy (patmaptableptr->Even, 
					numpats);
				bpi->GlobalPattern.Odd = GetCopy (patmaptableptr->Odd, 
					numpats);
			}
		}
#endif
		 else if (chunk.ChunkLabel == CHUNK_BATT)	{ /** B A T T **/
			
			/** Allocate Attributes Array **/
			if ( (bpi->AttribTable = (ULONG *)MEM_malloc (
				(size_t)chunk.ChunkSize)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadBPI:OOM allocating attribute tables:BATT(1) '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
			
			/** Read In Attributes **/
			if ((FPF_read (bpi->AttribTable, (size_t)chunk.ChunkSize)) 
				!= chunk.ChunkSize) 
			{
				GEcatf ("\nLoadBPI:Error reading BATT(2) Chunk");
				goto CLOSE_EXIT;
			}
		} else if (chunk.ChunkLabel == CHUNK_BINF)	{ /** B I N F **/

			/** Allocate Brush Info Array **/
			if ( (sizeinfo = (ULONG *)MEM_malloc (
				(size_t)chunk.ChunkSize)) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadBPI:OOM allocating size tables:BINF(1) '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
			
			/** Read In Brush Info **/
			if ((FPF_read (sizeinfo, (size_t)chunk.ChunkSize)) 
				!= chunk.ChunkSize) 
			{
				GEcatf ("\nLoadBPI:Error reading BINF(2) Chunk");
				goto CLOSE_EXIT;
			}
			bpi->Shapes = (GShape **)sizeinfo;
		} else if (chunk.ChunkLabel == CHUNK_BDAT)	{ /** B D A T **/
			GotBDAT = TRUE;
			/** Get Shape Processing Buffers **/
			if ( (RawBuf = (UBYTE *)GetGShapeLoadBuf ()) == NULL)
			{
				SetGlobalErr ( ERR_OUT_OF_MEMORY);
				GEprintf2 ("LoadBPI:OOM allocating RawBuff:BDAT(1) '%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}

			/** Load and process Brushes **/
			if (!LoadnTranslate (bpi, &bpi_info, numshapes)) 
			{
				goto CLOSE_EXIT;
			}	
		} else { /* Unknown or UnNeeded Chunk, Skip It */
			if (! FPF_skip (chunk.ChunkSize)) {
				SetGlobalErr ( ERR_READING_FILE);
				GEprintf2 ("LoadBPI:Error unknown chunk type/size:'%s'-%d", filename, fpfindex);
				goto CLOSE_EXIT;
			}
		}
	}

#if __MSDOS__
	if (foundMPAL && 
		(GraphicsMode == MODE_EGA || GraphicsMode == MODE_TANDY) &&
		!bpi->Palette.Data.EGA) {
		if ((bpi->Palette.Data.EGA = MEM_malloc (16)) == NULL) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEprintf2 ("LoadBPI:OOM allocating EGA fake palette:'%s'-%d", filename, fpfindex);
			goto CLOSE_EXIT;
		}
		for (temp = 0; temp < 16; temp++) {
			bpi->Palette.Data.EGA[temp] = temp;
		}
		bpi->Palette.NumColors = 16;
	}
#endif

#if __MSDOS__
	if (spacetoalloc)		DestroyTempBuffer (patdatptr, spacetoalloc);
	if (pattabarraysize)	DestroyTempBuffer (patmaptableptr, pattabarraysize);
#endif

	if (remapbuf2)			DestroyTempBuffer (ubyteptr, remapbuf2);
	if (remapbuf1)			DestroyTempBuffer (bpi_info.RemapInfo.Tables, remapbuf1);

	FPF_close ();
	return bpi;

CLOSE_EXIT:
	FPF_close ();
FREE_EXIT:
#if __MSDOS__
	if (spacetoalloc)		DestroyTempBuffer (patdatptr, spacetoalloc);
	if (pattabarraysize)	DestroyTempBuffer (patmaptableptr, pattabarraysize);
#endif

	if (remapbuf2)			DestroyTempBuffer (ubyteptr, remapbuf2);
	if (remapbuf1)			DestroyTempBuffer (bpi_info.RemapInfo.Tables, remapbuf1);

	FreeBPI (bpi);
	return NULL;
} /* LoadBPI */

/*********************************************************************
 *
 * FreeBPI
 *
 * SYNOPSIS
 *		void	FreeBPI (
 *			BPIType	*bpi
 *		)
 *
 * PURPOSE
 *		To free the memory allocated by LoadBPI.
 *
 * INPUT
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void	FreeBPI (
	BPIType	*bpi
)
{
	UWORD i;
	ULONG	attrib;
	
	if (!bpi)
		return;

	/** Free Shapes and Shape Table **/
	for (i = 0; i < bpi->NumShapes; i++) {
		attrib = bpi->AttribTable[i];
		if (! IsKlone(attrib)) { 
			switch ((short)UserCode (attrib)) {
			case DONT_TOUCH:
				MEM_free (bpi->Shapes[i]);
				break;
			case MAKE_GSHAPE:
				FreeGShape (bpi->Shapes[i]);
				break;
#if ERRORS
			default:
				printf ("bad shape type %d\n", (short)UserCode (attrib));
				break;
#endif
			}
		}
	}
	if (bpi->Shapes)
		MEM_free (bpi->Shapes);

	/** Free Attributes Table **/
	if (bpi->AttribTable) {
		MEM_free (bpi->AttribTable);
	}

	/** Free any global pattern **/
#if __MSDOS__
	if (bpi->GlobalPattern.NumPatterns) {
		MEM_free (bpi->GlobalPattern.Odd);
		MEM_free (bpi->GlobalPattern.Even);
	}
#endif

	/** Free Palette **/
	if (bpi->Palette.NumColors) {
#if __MSDOS__
		switch (GraphicsMode) {
		case MODE_MCGA:
			MEM_free (bpi->Palette.Data.RGB);
			break;
		case MODE_EGA: case MODE_TANDY:
			MEM_free (bpi->Palette.Data.EGA);
			break;
		}
#endif
#if __AMIGAOS__
#endif
	}

	if (bpi->PaintBox) {
		FreePBX (bpi->PaintBox);
	}

	MEM_free (bpi);
} /* FreeBPI */

