/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * FPFLoad.C
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 02/13/90
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Routines for loading from a FilePack File (FPF).  An FPF file is
 *		a collection of LZSS compressed files preceded by a table of offsets
 *		to the compressed sub-units.
 *		The actual decompression code was pulled from a public domain program 
 *		by Haruhiko Okumura which was modified by Danial Chang (Separated 
 *		Decode() into InitDecode() and repeated calls to lzssgetc()).
 *		Many thanks to both.
 *	
 *		Header comments from original public domain LZSS program:
 *			**************************************************************
 *				LZSS.C -- A Data Compression Program
 *				(tab = 4 spaces)
 *			***************************************************************
 *				4/6/1989 Haruhiko Okumura
 *				Use, distribute, and modify this program freely.
 *				Please send me your improved versions.
 *					PC-VAN		SCIENCE
 *					NIFTY-Serve	PAF01022
 *					CompuServe	74050,1022
 *			**************************************************************
 *
 *			Modified by Daniel Chang, 08/15/89. Separated Decode()
 *				into InitDecode() and repeated calls to lzssgetc(). 
 *	
 * HISTORY
 *	05/05/90 Saturday (JMA) - Made FPF_open() use user routine to make sure
 *			file exists.
 *	
 *	08/09/90 Thursday (JMA) - Made FPF routines preserve errno and made
 *			them return -1 on error just like the Unix file routines 
 *			they replace.
 *
 *	05/29/91 Wednesday (RGM) - New SAS/Lattice Amiga library version changes.
 *
 *	08/26/92 Wednesday (GAT) -
 *	
 *			1) Added support for cached FPF entries by making local open,
 *			   read, close routines.
 *	
 *			2) Switched	to EIO routines to make it more portable.
 *	
 *			3) Moved fstrpcmp to echidna\strings because it is more
 *			   globally useful.
 *	
 *			NOTE: I see no point in supporting the ERRNO
 *			variable of the Unix file routines since it is ILLEGAL to
 *			to use the Unix file routines in Echidna programs.
 *			See 'eggrules.txt'
 *	
 *			Note: I also believe this file should be modified to be
 *			a library file with MODULE statements like 'eio.c' so that
 *			the FPF_findnext and other directory stuff is not linked in
 *			because they are a waste of memory if they are not being
 *			used and they are not used in most echidna programs.
 *	
 *			4) Separated the FPF directory stuff.
 *	
 *			5) Removed the #if ERROR, FPFName = filename; lines from
 *			   FPF_GetDir to simplify separation of the directory stuff.
 *			   If it is really needed then FPFName would need to be global
 *			   or switch to C++ and make it a private static data member.
 *	
 *	08/27/92 Thursday (GAT) - 
 *	
 *			1) Added Resource awareness.  Not a perfect implementation :-<
 *	
 *			2) Didn't finish it because what's the point.  The user can
 *			   run SMARTDRV or something else and probably get the same
 *			   benefit.  The bottleneck in the FPF system is current
 *			   the Egetc and lzss stuff being in 'C' instead of assembly.
 *			   I think adding the cache wouldn't really speed it up more
 *			   than 5% to 10% whereas speeding up the lzss/egetc stuff
 *			   could speed it up 80% to 150% or more.
 *	
 *			   So, should I remove the cache support?  Yes with
 *			   #define USE_CACHE	0
 *	
 *	
 *	
*/
#include <echidna/platform.h>
#include "switches.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#if LATTICE
#include <errno.h>
#endif

#include <echidna/etypes.h>
#include <echidna/eio.h>
#include <echidna/strings.h>
#include <echidna/fpfload.h>
#include <echidna/eerrors.h>
#include <echidna/memory.h>
#include <echidna/resource.h>

/**************************** C O N S T A N T S ***************************/

#define N		 4096	/* size of ring buffer */
#define F		   18	/* upper limit for match_length */
#define THRESHOLD	2   /* Encode string into position and length */
						/*   if match_length is greater than this */
#define TEXT_BUF_SIZE	(N + F - 1)

#define LZSS_INIT_CHAR	' '	/* char to initialize ring buffer with */

#define FBUF_SIZE	512

#define INDIRECT_BIT_MASK		(1L << 31)
#define UNCOMPRESSED_BIT_MASK	(1L << 30)
#define OFFSET_MASK			(~(UNCOMPRESSED_BIT_MASK | INDIRECT_BIT_MASK))

#define	USE_CACHE	0

/******************************** T Y P E S *******************************/

typedef struct {
	short			DirIndex;
	unsigned short	FPF_Index;
} DirIndexesType;

typedef struct {
	DirIndexesType	*DirIndexes;
	char			*FilenameBase;
} FPFDirTableType;

typedef struct {
	short			NumFiles;
	short			NumDirFiles;
	unsigned short	SizeOfDirTable;
} FPFHeaderType;

/*------------------------------------------------------------------------*/
/**# MODULE:FPF_Open                                                      */
/*------------------------------------------------------------------------*/

/****************************** G L O B A L S *****************************/

static unsigned char	*fbuf = NULL;
static unsigned char	*fbufptr;
static long				 fchrs;

static unsigned char	*text_buf = NULL;	/* ring buffer of size N, */
											/* with extra F-1 bytes to */
											/* facilitate string comparison */

static int	 infile;

/* LZSS decompression counters and flags. */
static short i;
static short j;
static short k;
static short r;
static unsigned short flags;

/* FPF counters and flags */
static short	External;		/* True if actual data not in an FPF. */
static short	Uncompressed;	/* True if data is uncompressed. */
static long		FPFOffset;		/* Current offset in data file */
static long		iCount;			/* Char counter and Egetc error flag 
									for FPF_read() */
static long		Count;	/* Char counter and Egetc error flag 
									for FPF_skip() */
static long		BytesLeft;		/* Bytes of data remaining to be read. */

static FPFHeaderType	FPFHeader;


//
// Resource Manager sets up this point before calling FPF_open
// so that FPF_open can use cached resources.  This variable is cleared
// after each use so that the other calls FPF_open will be safe.
//
// This var is set up thought FPF_SpecifiyCache
//
#if USE_CACHE
static ResourceTable	*FPFResource;
static ResourceCache	*FPFRSCache;
static long				 RSPosition;
#endif

#if ERRORS
static char		*FPFName;
static short	 FPFIndex;
#endif

/*	'OpenFileName' contains the name of the file currently opened
	by FPF open routines.  It may be an FPF file or an 'External' data file.
	Check 'External' flag to determine which it is.  Contains empty string
	if no file is currently open.
*/                                   
static char	  OpenFileName[EIO_MAXPATH];

FileExistsFuncPtr	FPF_FileExistsFunc = EIO_FileExists;

/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/

#if USE_CACHE
static int fpf_close (int fh)
{
	if (!FPFRSCache && fh >= 0)
	{
		return EIO_Close (fh);
	}
	return 0;
}
static long fpf_read (int fh, void *fbuf, long size)
{
	if (FPFRSCache)
	{
		UBYTE huge32 *mem;
		long		 bytes;

		mem   = (UBYTE huge32 *)fbuf;
		bytes = size;

		while (bytes)
		{
			short	xtra;
			UWORD	offset;
			UWORD	len;

			xtra   = RSPostion / RS_MAX_BYTES_PER_XTRA;
			offset = RSPostion % RS_MAX_BYTES_PER_XTRA;
			len    = RS_MAX_BYTES_PER_XTRA - offset;
			len    = min (len, bytes);

			ReadXTRA (FPFRSCache->Xtra[xtra], offset, len, mem);

			mem        += len;
			bytes      -= len;
			RSPosition += len;
		}

		return size;
	} else {
		return EIO_Read (fh, fbuf, size);
	}
}

//
// note: only one type is used, EIO_SEEK_CURRENT so that's all I implemented
//
static int fpf_seek (int fh, long pos, short type)
{
	if (FPFRSCache)
	{
		RSPosition += pos;
		return RSPosition;
	} else {
		return EIO_Seek (fh, pos, type);
	}
}
#else
#define	fpf_seek	EIO_Seek
#define	fpf_read	EIO_Read
#define	fpf_close	EIO_Close
#endif

/*********************************************************************
 *
 * FPF_SetFileExistsFunc
 *
 * SYNOPSIS
 *		void FPF_SetFileExistsFunc (
 *			FileExistsFuncPtr	fefptr
 *		)
 *
 * PURPOSE
 *		To set the function used by the FPF routines when they
 *		check to see if a file exists before opening it.  If this
 *		function is not set then no check will be made before opening.
 *
 *		If your application has data on multiple disks, then you can
 *		write a 'FileExists' function that will make sure the right
 *		disk is in the drive (i.e. prompt the user to swap disks if
 *		necessary).
 *
 * INPUT
 *		fefptr:	Pointer to function which takes as an argument a 
 *				pointer to a file name string, and returns a short 
 *				int whose value has the meaing:
 *					0			File does NOT exit.
 *					non-zero	File Exists.
 *
 * EFFECTS
 *		None.
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		'FileExistsFuncPtr' typedef in FPFLoad.H
*/
void FPF_SetFileExistsFunc (
	FileExistsFuncPtr	fefptr
)
{
	FPF_FileExistsFunc = fefptr;

} /* FPF_SetFileExistsFunc */

/*********************************************************************
 *
 * fillfbuf
 *
 * SYNOPSIS
 *		static short fillfbuf (void)
 *
 * PURPOSE
 *		Used by Egetc() to fill in Egetc buffer with characters from 
 *		input file.
 *		Egetc() is a macro that gets characters from the Egetc buffer
 *		until there are none left, at which point it calls fillfbuf()
 *		to fill the buffer with more characters from the input file.
 *		This operation is transparent to users of Egetc().
 *
 * INPUT
 *		None.
 *
 * EFFECTS
 *		Will set Count and iCount to read() result on error.  These
 *		are flags for FPF_skip() and FPF_read() (respectively) that
 *		allow them to return error value.
 *
 * RETURN VALUE
 *		Returns next charater from file.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
#define Egetc(infile)	\
	((fchrs-- == 0) ? (fillfbuf()) : (*fbufptr++))

static short fillfbuf (void)
{
	fchrs = fpf_read (infile, fbuf, FBUF_SIZE);
	if (fchrs == -1 || fchrs == 0) {
		Count = iCount = fchrs; /* may be 0 or -1 */
		return EOF;	
	}
	fbufptr = fbuf;
	fchrs--;
	return *fbufptr++;
	
} /* fillfbuf */

/*********************************************************************
 *
 * InitDecode
 *
 * SYNOPSIS
 *		void InitDecode(void)
 *
 * PURPOSE
 *		To initialize the 4k ring buffer, flags and counters for 
 *		LZSS decompression.
 *
*/
void InitDecode(void)
{
	memset (text_buf, LZSS_INIT_CHAR, (N - F));
	r = N - F;  
	flags = 0;
	k = 1; j = 0;			/* make sure k<=j is initially false. */
} /* InitDecode */

/*********************************************************************
 *
 * FPF_read
 *
 * SYNOPSIS
 *		long	FPF_read (
 *			void	*buffer, 
 *			long	 len
 *		)
 *
 * PURPOSE
 *		Read from an FPF file.
 *
 * INPUT
 *		buffer:	Where to put bytes.
 *		len:	How many to put.
 *
 * RETURN VALUE
 *		On successful completion , FPF_read returns the number of
 *		bytes placed in the buffer.
 *		Warning! Will read more bytes than are in file if you ask
 *		for them.
 *		On end-of-file, FPF_read returns 0.  
 *		On error, FPF_read returns -1 and errno has the value
 *		returned by the UNIX function read(), which is used by FPF_read.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		FPF_open(), FPF_openf(), FPF_skip(), FPF_close()
*/
long FPF_read (
	void	*buffer, 
	long	 len
)
{
	short		c;
	
	if (BytesLeft <= 0) {
		return EOF;
	}
	if (Uncompressed) {
		iCount = fpf_read (infile, buffer, len);
	} else {
		char huge32 *mem = (char huge32 *)buffer;

		iCount = 0;
		while (iCount < len) {
			if (k <= j) {			/* continue outputting characters? */
				c = text_buf[(i + k) & (N - 1)];
				text_buf[r++] = c;  r &= (N - 1);
				k++;
				*mem++ = (char)c;
				iCount++;
			} else {
				if (((flags >>= 1) & 256) == 0) {
					if ((c = Egetc(infile)) == EOF) goto DONE;
					flags = c | 0xff00;		/* uses higher byte cleverly */
				}									/* to count eight */
				if (flags & 1) {
					if ((c = Egetc(infile)) == EOF) goto DONE;
					text_buf[r++] = c;  r &= (N - 1);
					*mem++ = (char)c;
					iCount++;
				} else {
					if ((i = Egetc(infile)) == EOF) goto DONE;
					if ((j = Egetc(infile)) == EOF) goto DONE;
					i |= ((j & 0xf0) << 4);  j = (j & 0x0f) + THRESHOLD;

					/* j is guaranteed >= 0, code below runs at least once. */

					k = 0;
					c = text_buf[(i + k) & (N - 1)];
					text_buf[r++] = c;  r &= (N - 1);
					k++;
					*mem++ = (char)c;
					iCount++;
				}
			}
		}
	}

DONE:
	if (iCount != -1L) {
		FPFOffset += iCount;
		BytesLeft -= iCount;
#if ERRORS
		if (iCount != len) {
			SetGlobalErr ( ERR_READING_FILE);
			GEprintf1 ("FPF_read: %s", GlobalErrStr[ERR_READING_FILE]);
			GEcatf1 (" '%s'", FPFName);
			if (External) {
				GEcatf1 ("[%s]", OpenFileName);
			} else {
				GEcatf1 ("-%d", FPFIndex);
			}
		}
#endif
	}
	return iCount;

} /* FPF_read */

/*********************************************************************
 *
 * FPF_skip
 *
 * SYNOPSIS
 *		long FPF_skip (
 *			long	offset
 *		)
 *
 * PURPOSE
 *		To skip 'offset' number of bytes forward in FPF data file.
 *
 * INPUT
 *		offset:	Number of bytes to skip from current position.
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		FPF_skip returns the offset of the pointer's new position
 *		measured in bytes from the file beginning.  FPF_skip returns
 *		-1L on error and errno has the value returned by the UNIX 
 *		function read(), which is used by FPF_skip.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		FPF_open(), FPF_openf(), FPF_read(), FPF_close()
 *		
*/
long FPF_skip (
	long	offset
)
{
	short		c;

	Count = 0;

	if (BytesLeft <= 0) {
		return FPFOffset;
	}
	if (Uncompressed) {
		if (fpf_seek (infile, offset, EIO_SEEK_CURRENT) != (-1)) { /* success */
			Count = offset;
		} else {
			Count = -1L;
		}
	} else {
		while (Count < offset) {
			if (k <= j) {			/* continue outputting characters? */
				c = text_buf[(i + k) & (N - 1)];
				text_buf[r++] = c;  r &= (N - 1);
				k++;
				Count++;
			} else {
				if (((flags >>= 1) & 256) == 0) {
					if ((c = Egetc(infile)) == EOF) goto DONE;
					flags = c | 0xff00;		/* uses higher byte cleverly */
				}									/* to count eight */
				if (flags & 1) {
					if ((c = Egetc(infile)) == EOF) goto DONE;
					text_buf[r++] = c;  r &= (N - 1);
					Count++;
				} else {
					if ((i = Egetc(infile)) == EOF) goto DONE;
					if ((j = Egetc(infile)) == EOF) goto DONE;
					i |= ((j & 0xf0) << 4);  j = (j & 0x0f) + THRESHOLD;

					/* j is guaranteed >= 0, code below runs at least once. */

					k = 0;
					c = text_buf[(i + k) & (N - 1)];
					text_buf[r++] = c;  r &= (N - 1);
					k++;
					Count++;
				}
			}
		}
	}

DONE:
	FPFOffset += Count;
	BytesLeft -= Count;
#if ERRORS
	if (Count != offset) {
		SetGlobalErr ( ERR_READING_FILE);
		GEprintf1 ("FPF_read: %s", GlobalErrStr[ERR_READING_FILE]);
		GEcatf1 (" '%s'", FPFName);
		if (External) {
			GEcatf1 ("[%s]", OpenFileName);
		} else {
			GEcatf1 ("-%d", FPFIndex);
		}
	}
#endif
	return Count;

} /* FPF_skip */

/*********************************************************************
 *
 * FPF_open
 *
 * SYNOPSIS
 *		int FPF_open (
 *			char	*filename
 *			short	 selector
 *		)
 *
 * PURPOSE
 *		To setup access to a file containing data which may be compressed 
 *		or uncompressed and do whatever preparation is necessary for
 *		the rest of the FPF routines (FPF_read, FPF_skip, FPF_close)
 *		to function properly on the data and on the file.  
 *
 *		The FPF routines support 3 different mechanisms for accessing 
 *		data:
 *			Non-FPF
 *			FPF
 *			FPF-Indirect.  
 *
 *		Data files accessed through Non-FPF and FPF-Indirect methods are
 *		considered 'External', meaning that the actual data is not inside 
 *		an FPF file.
 *		In each mechanism the data may be compressed or
 *		uncompressed which means there are actually 6 different
 *		ways of accessing data.  Which mechanism is used and whether
 *		the data is treated as compressed or uncompressed depends
 *		upon the value of 'selector' or upon the contents of the file 
 *		named by 'filename'.  
 *
 *		Non-FPF:
 *		If 'selector' is negative, then 'filename' is not an FPF file 
 *		but is a data file containg compressed or uncompressed data as 
 *		the following table indicates:
 *
 *			selector	Compressed	
 *			----------	----------	
 *			-2			Yes			
 *							
 *			-1			No			
 *
 *		Behavior is undefined for other negative values.
 *
 *		FPF and FPF-Indirect:
 *		If 'selector' is 0 or greater then 'filename' is an FPF file and 
 *		'selector' is used to index a table within the FPF that tells whether
 *		the data is compressed or uncompressed and where the data is (i.e.
 *		at what offset in the FPF or in what external file).  The contents
 *		of this table is determined by the instructions in the FPS file
 *		and any command line arguments to FilePack when the FPF was
 *		created.
 *
 *		The table entry is a long with information in the two most 
 *		significant bits regarding the access mechanism and compression.  
 *		Bit 31 determines the mechanism and bit 30 determines compression 
 *		status:
 *
 *		MS Bits	Mechanism		Compressed	Data Location
 *		-------	---------		----------	-------------
 *		00		FPF				Yes			In FPF file.
 *		01		FPF				No			In FPF file.
 *		10		FPF-Indirect	Yes			In another file.
 *		11		FPF-Indirect	No			In another file.
 *
 *		When the two most significant bits are masked off, the remaining
 *		value is the offset in the FPF file to either the start of the
 *		data or the start of a null terminated string containing the name
 *		of an external file which contains the data.
 *
 *
 * INPUT
 *		filename:	String containing name of a data file or of FPF.
 *		selector:	Must be >= -2.  See discussion under PURPOSE above.
 *
 * EFFECTS
 *		Memory will be allocated for a file buffer and decompression
 *		buffer if the data is compressed.  This memory will be freed
 *		when FPF_close() is called.
 *		
 *		WARNING: This routine is not re-entrant.  Only one file may be
 *			open at at time.  Once you open a file with FPF_open() or
 *			FPF_openf(), you must call FPF_close() before you can use
 *			either routine to open another file.
 *
 * RETURN VALUE read
 *		On success FPF_open returns a non-negative integer. 
 *		On error FPF_open returns -1 and errno has the value returned by 
 *		the UNIX function open(), which is used by FPF_open.
 *
 *		Call FPF_close() to close a file opened with this routine.
 *	
 * HISTORY
 *	08/10/90 Friday (JMA) - Modified to handle non FPF files, uncompressed
 *			FPF data and compressed indirect files.
 *		
 *	08/27/90 Monday (JMA) - Modified to handle new format containing 
 *			filename directory information.
 *
 * SEE ALSO
 *	FPF_openf(), FPF_read(), FPF_skip(), FPF_close()
 *
*/
int FPF_open (
	char	*filename,
	short	 selector
)
{
	long	offset;
	int		myerrno;
	int		indirect;

#if ERRORS
	FPFName = filename;
	FPFIndex = selector;
#endif

	External = TRUE;	/* Assume this because it less code */

	/* Make sure buffers dont get freed on error */
	text_buf = fbuf = NULL;

	#if USE_CACHE
	/* Clear last cache setup */
	FPFRSCache = NULL;

	//
	// If and FPFResource is setup then see if selector is cached
	//
	if (FPFResource &&
		selector - FPFResource->FPF_Index >= 0 &&
		selector - FPFResource->FPF_Index < FPFResource->CacheCount)
	{
		ResourceCache	*rc;

		selector -= FPFResource->FPF_Index
		rc = FPFResource->FirstCache;

		while (selector)
		{
			rc = ((UBYTE *)rc) + sizeof (ResourceCache) + 
				 (rc->XtraCount - 1) * sizeof (XTRAPntr);

			selector--;
		}

		Indirect	 = FALSE;
		Uncompressed = rc->Flags & UNCOMPRESSED_BIT_MASK;
		BytesLeft    = rc->Size;
		RSPosition   = 0;
		FPFRSCache   = rc;
		infile		 = 0;

		FPFResource  = NULL;

	}
	else
	#endif
	{

		/* Check if user supplied a 'file exists' function */
		if (! FPF_FileExistsFunc (filename)) {
			return -1;
		}

		/* Open file */
		if ((infile = EIO_ReadOpen (filename)) == -1) {
			SetGlobalErr ( ERR_FILE_NOT_FOUND);
			GEprintf1 ("FPF_Open: %s", GlobalErrStr[ERR_FILE_NOT_FOUND]);
			GEcatf1 (" '%s'", filename);
			return -1;
		}
	
		/* Save filename */
		strcpy (OpenFileName, filename);

		/* Check if it is Non-FPF file */
		if (selector < 0) { 
			/* Mechanism is Non_FPF */
			Uncompressed = (selector == -1);
		} else {

			/* It is an FPF file. */
		
			/* Read Header info */
			if (EIO_Read (infile, &FPFHeader, sizeof (FPFHeader)) != 
				sizeof (FPFHeader))
			{
				goto CLOSE_FAIL;
			}

			/* Check that FP index is in range */
			if (selector >= FPFHeader.NumFiles) {
				goto CLOSE_FAIL;
			}

			/* Calculate offset to selected entry in Offset Table */
			offset = sizeof (FPFHeader) +
				FPFHeader.NumDirFiles * sizeof (DirIndexesType) + 
				FPFHeader.SizeOfDirTable + 
				selector * sizeof (long);

			/* Seek to selected Offset table entry */
			if (EIO_Seek (infile, offset, EIO_SEEK_BEGINNING) == -1) {
				goto CLOSE_FAIL;
			}

			/* Read Offset to Data */
			if (EIO_Read (infile, &offset, sizeof (long)) == -1) {
				goto CLOSE_FAIL;
			}

			Uncompressed = ((offset & UNCOMPRESSED_BIT_MASK) != 0);
			indirect = ((offset & INDIRECT_BIT_MASK) != 0);

			/* Seek to Start of Data or filename */
			offset &= OFFSET_MASK;
			if (EIO_Seek (infile, offset, EIO_SEEK_BEGINNING) == -1) {
				goto CLOSE_FAIL;
			}

			if (indirect) {	/* Mechanism is FPF-Indirect */
				/* Get name of external file */
				if (EIO_Read (infile, OpenFileName, EIO_MAXPATH) <= 0) {
					goto CLOSE_FAIL;
				} 

				/* Close FPF file */
				if (EIO_Close (infile) < 0) {
					goto CLOSE_FAIL;
				}

				/* Open external file */
				if ((infile = EIO_ReadOpen (OpenFileName)) == -1) {
					SetGlobalErr ( ERR_READING_FILE);
					GEprintf2 ("FPF_Open:Opening(4) '%s' ['%s']", filename, OpenFileName);
				} 
			} else  { /* mechanism is FPF */
				External = FALSE;
				/* Read Data size long word */
				if (EIO_Read (infile, &BytesLeft, sizeof (long)) != sizeof (long)) {
					goto CLOSE_FAIL;
				}
			}
		}

		/* Get file length if file is External */
		if (External) {
			if ((BytesLeft = EIO_FileLength (infile)) == -1) { 
				goto CLOSE_FAIL;
			}
		}
	}

	if (!Uncompressed && infile >= 0) {

		/* Initialize and allocate Egetc() input buffer */
		fchrs = 0;
		if ( (fbuf = CreateTempBuffer (FBUF_SIZE)) == NULL) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEprintf ("OOM: Egetc Buffer");
			goto CLOSE_FAIL;
		}

		/* Allocate buffer for LZSS decompression */
		if ( (text_buf = CreateTempBuffer (TEXT_BUF_SIZE)) == NULL) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEprintf ("OOM: LZSS Buffer");
			goto CLOSE_FAIL;
		}

		/* Initialize Decompressor */
		InitDecode ();
	}

	/* Initialize offset within data file */
	FPFOffset = 0; 

	return infile;

CLOSE_FAIL:
	myerrno = errno;
	SetGlobalErr ( ERR_READING_FILE);
	GEcatf2 ("FPF_Open: Opening(5) '%s'-%d", FPFName, FPFIndex);
	FPF_close ();
	errno = myerrno;

	return -1;

} /* FPF_open */

/*********************************************************************
 *
 * FPF_close
 *
 * SYNOPSIS
 *		int	FPF_close (void)
 *
 * PURPOSE
 *		On success FPF_close returns 0.
 *		On error FPF_close returns -1 and errno will have the value 
 *		set by the function close, which is	used by FPF_close.
 *
 * INPUT
 *		None.
 *
 * EFFECTS
 *		None.
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		FPF_open(), FPF_openf()		
*/
int	FPF_close (void)
{

	OpenFileName[0] = '\0';

	if (text_buf) 
        DestroyTempBuffer (text_buf, TEXT_BUF_SIZE);

	if (fbuf) 
        DestroyTempBuffer (fbuf, FBUF_SIZE);

	return fpf_close (infile);

} /* FPF_close */

/*********************************************************************
 *
 * FPF_BytesLeft
 *
 * SYNOPSIS
 *		long FPF_BytesLeft (void)
 *
 * PURPOSE
 *		To determine the number of bytes remaining to be read in
 *		a data file opened by and FPF open routine.
 *
 * INPUT
 *		None.
 *
 * EFFECTS
 *		None.
 *
 * RETURN VALUE
 *		Returns Number of bytes remaining to be read in data file.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long FPF_BytesLeft (void)
{
	return BytesLeft;
} /* FPF_BytesLeft */


/*------------------------------------------------------------------------*/
/**# MODULE:FPF_DeadOrAlive                                               */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * FPF_openDeadOrAlive
 *
 * SYNOPSIS
 *		int FPF_openDeadOrAlive (
 *			char	*FPFfilename,
 *			short	 selector,
 *			char	*DataFilename,
 *			short   *pfUsedFPF
 *		)
 *
 * PURPOSE
 *		To see if DataFilename exists in the current directory and open
 *		it if it does for FPF reading. If it does not exist then open the
 *		file indexed by selector in the FPF file FPFfilename.  Fill in
 *		flag that tells if the FPF file was used.
 *		
 *
 * INPUT
 *		FPFfilename:	name of FPF file.
 *		selector:		FPF index.
 *		DataFilename:	File that you prefer to open if it exists.
 *		pfUsedFPF:		Pointer to flag to set if DataFilename was not found
 *						and the FPF file was used instead. May be NULL.
 *	
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		See FPF_open return values.
 *		*pfUsedFPF is not modified if the open fails both cases.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
int FPF_openDeadOrAlive (
	char	*FPFfilename,
	short	 selector,
	char	*DataFilename,
	short	*pfUsedFPF
)
{
	int result;

	if ((result = FPF_open (DataFilename, -1)) != -1) {
		if (pfUsedFPF) {
			*pfUsedFPF = FALSE;
		}
	} else if ((result = FPF_open (FPFfilename, selector)) != -1) { 
		if (pfUsedFPF) {
			*pfUsedFPF = TRUE;
		}
	}

	return result;

} /* FPF_openDeadOrAlive */

/*********************************************************************
 *
 * FPF_RecreateSourceFile
 *
 * SYNOPSIS
 *		short FPF_RecreateSourceFile (
 *			char	*pszFpfFilename, 
 *			short	 selector, 
 *			char	*pszDataFilename
 *		)
 *
 * PURPOSE
 *		To recreate the source file from the FPF compressed file under
 *		the given name.
 *
 * INPUT
 *		pszFpfFilename:	FPF file name.	Non-NULL.
 *		selector:		FPF index of packed file for which to recreate the
 *						source file. 
 *		pszDataFilename: Name to give new source file. Non-NULL.
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		Returns TRUE on success.
 *		Returns FALSE on error. 
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short FPF_RecreateSourceFile (
	char	*pszFpfFilename, 
	short	 selector, 
	char	*pszDataFilename
)
{
#define BufSize		512
	char Buf[BufSize];
	short fResult;
	
	if ( (fResult = (-1 != FPF_open (pszFpfFilename, selector))) != 0) {
		int hndl;
		long lsizeLeft = FPF_BytesLeft();

		if ( (fResult = (-1 != (hndl = EIO_WriteOpen (pszDataFilename)))) != 0) {
			long lsizeRead;

			lsizeRead = min (lsizeLeft, BufSize);
			while ( (fResult = (lsizeRead == FPF_read(Buf, lsizeRead))) != 0)
			{
				if (!(fResult = (-1 != EIO_Write(hndl, Buf, lsizeRead)))) {
					break;
				}
				if (!(lsizeLeft -= lsizeRead)) {
					break;
				}	
				lsizeRead = min (lsizeLeft, BufSize);
			}
			EIO_Close (hndl);
		}

		FPF_close ();
	}

	return fResult;

} /* FPF_RecreateSourceFile */


/*------------------------------------------------------------------------*/
/**# MODULE:FPF_Directory                                                 */
/*------------------------------------------------------------------------*/

static DirIndexesType	*Dir_Indexes;
static unsigned short GetDirSize;

/*********************************************************************
 *
 * FPF_FreeDir
 *
 * SYNOPSIS
 *		void	FPF_FreeDir (void)
 *
 * PURPOSE
 *		To free memory allocated by FPF_GetDir().
 *
 * INPUT
 *		dirtable:	Non-NULL pointer returned by FPF_GetDir().
 *
 * EFFECTS
 *		Frees memory.
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		FPF_GetDir()
*/
void	FPF_FreeDir (void)
{
    DestroyTempBuffer (Dir_Indexes, GetDirSize);

} /* FPF_FreeDir */

/*********************************************************************
 *
 * FPF_GetDir
 *
 * SYNOPSIS
 *		short	FPF_GetDir (
 *			char			*filename,
 *			FPFHeaderType	*fpfheader,
 *			FPFDirTableType	*dirtable
 *		)
 *
 * PURPOSE
 *		To read in the header and directory table of an FPF.
 *
 * INPUT
 *		filename:	FPF name.
 *		fpfheader:	Pointer to FPFHeaderType struct to fill out. Pass
 *					NULL if you do not want this information.
 *		dirtable:	Pointer to FPFDirTableType struct to fill out.  
 *					NULL if you do not want this information.
 *
 * EFFECTS
 *		Fills in 'fpfheader' if 'fpfheader is non-NULL.
 *		Allocates memory for directory info and fills out 'dirtable if
 *		'dirtable' is non-NULL.
 *		
 *		This routine may be called on an FPF that is open or closed. 
 *		Note: An FPF does not actually remain open after the open call
 *			if the data file requested was 'External' to the FPF.
 *		Warning: If the FPF was already open and this routine gets an
 *			error, then it will close the FPF.
 *
 * RETURN VALUE
 *		On success returns non-zero.  Use FPF_FreeDir() to 
 *		deallocate memory allocated by FPF_GetDir() if 'dirtable' was
 *		requested.
 *		On failure returns 0 (do not call FPF_FreeDir() on failure).
 *
 * HISTORY
 *	
 *	08/26/92 Wednesday (GAT) - Why did this routine use the open FPF
 *		file ever?  I changed it so it just opens the file reguardless
 *		and it do NOT close FPF on failure!
 *
 *
 * SEE ALSO
 *		FPF_FreeDir()
*/
short	FPF_GetDir (
	char			*filename,
	FPFHeaderType	*fpfheader,
	FPFDirTableType	*dirtable
)
{
	int				 myerrno;
	int				 myhndl;
	FPFHeaderType	 fpf_header;
	char			*filename_base;
#if 0
	#if ERRORS
		FPFName  = filename;
	#endif
#endif

	Dir_Indexes = NULL;

	/* Check if user supplied a 'file exists' function */
	if (! FPF_FileExistsFunc (filename)) {
		return FALSE;
	}

	/* Open file */
	if ((myhndl = EIO_ReadOpen (filename )) == -1) {
		SetGlobalErr ( ERR_FILE_NOT_FOUND);
		GEprintf1 ("FPF_GetDir: %s", GlobalErrStr[ERR_FILE_NOT_FOUND]);
		GEcatf1 (" '%s'", filename);
		return FALSE;
	}

	/* Read Header info */
	if (EIO_Read (myhndl, &fpf_header, sizeof (fpf_header)) != 
		sizeof (fpf_header))
	{
		goto CLOSE_FAIL;
	}

	/* Copy Header if requested */
	if (fpfheader) {
		*fpfheader = fpf_header;
	}

	/* Get dirtable if requested */
	if (dirtable) {

		/* Calculate size of Directory structure */
		GetDirSize = fpf_header.NumDirFiles * sizeof (DirIndexesType) +
					fpf_header.SizeOfDirTable;

		/* Allocate buffer for Directory structure */
		if ((Dir_Indexes = (DirIndexesType *)CreateTempBuffer (GetDirSize)) == NULL) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEprintf ("OOM: Directory Buffer");
			goto CLOSE_FAIL;
		}

		/* If directory structure was packed then here I would have to:
				Allocate lzss buffer 
				Allocate Egetc buffer 
				Init Egetc 
				Init lzss buffer
				Assign BytesLeft
				Zero FPFOffset
				Call FPF_Read to read in the data.
				Call FPF_close to free the lzss and Egetc buffers and close 
					the file.
				Get rid of the current call to read (myhndl, . . .
				Get rid of the current call to close (myhndl)
		*/

		/* Read in Directory structure */
		if (EIO_Read (myhndl, Dir_Indexes, GetDirSize) != GetDirSize) {
			goto CLOSE_FAIL;
		}

		/* Calculate file name base */
		filename_base = (char *)Dir_Indexes + 
			(fpf_header.NumDirFiles * sizeof (DirIndexesType));

		/* Fill out dirtable */
		dirtable->DirIndexes = Dir_Indexes;
		dirtable->FilenameBase = filename_base;
	}

	EIO_Close (myhndl);

	return TRUE;

CLOSE_FAIL:
	myerrno = errno;
	if (Dir_Indexes)
	    DestroyTempBuffer (Dir_Indexes, GetDirSize);
	SetGlobalErr ( ERR_READING_FILE);
	GEcatf1 ("FPF_GetDir: Opening(5) '%s'", filename);
	EIO_Close (myhndl);
	errno = myerrno;

	return FALSE;
} /* FPF_GetDir */

/*********************************************************************
 *
 * FPF_openf
 *
 * SYNOPSIS
 *		int	FPF_openf (
 *			char	*filename,
 *			char	*datafilename,
 *			short	 selector
 *		)
 *
 * PURPOSE
 *		To get the FPF index for the given 'datafilename' from the
 *		given FPF file and then call FPF_open() to open it.
 *
 * INPUT
 *		filename:		FPF name.
 *		datafilename:	File name that appears in the directory table
 *						of the FPF file.
 *		
 *
 * EFFECTS
 *		WARNING: This routine is not re-entrant.  Only one file may be
 *			open at at time.  Once you open a file with FPF_open() or
 *			FPF_openf(), you must call FPF_close() before you can use
 *			either routine to open another file.
 *
 * RETURN VALUE
 *		On success FPF_open returns a non-negative integer. 
 *		On error FPF_openf returns -1 and errno has the value returned by 
 *		the UNIX function open(), which is used by FPF_openf.
 *
 *		Call FPF_close() to close a file opened with this routine.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		FPF_open(), FPF_read(), FPF_skip() FPF_close()
*/
int	FPF_openf (
	char	*filename,
	char	*datafilename,
	short	 selector
)
{
	struct FPF_tracker	fpftracker;	

	/* Check if filename is a Non-FPF file */
	if (selector < 0) {
		/* Mechanism is Non_FPF */
		return FPF_open (datafilename, selector);
	} 

	/* If data file in directory, then open it by FilePack index */
	if (FPF_findfirst (filename, datafilename, &fpftracker) >= 0) {
		return FPF_open (filename, fpftracker.FPF_IndexFound);
	}

	return -1;
} /* FPF_openf */

/*********************************************************************
 *
 * FPF_Info
 *
 * SYNOPSIS
 *		FPF_InfoType *FPF_Info (
 *			char 			*filename,
 *			FPF_InfoType	*fpf_info
 *		) 
 *
 * PURPOSE
 *		To get information about an FPF. The information is put
 *		in the 'fpf_info' structure:
 *			typedef struct FPF_Info {
 *				short	NumFiles;		
 *				short	NumDirFiles;	
 *			} FPF_InfoType;
 *
 * INPUT
 *		filename:	FPF name.
 *		fpf_info:	Pointer to structure to fill out.
 *
 * EFFECTS
 *		Fills in values in 'fpf_info'.
 *
 *		This routine may be called on an FPF that is open or closed. 
 *		Note: An FPF does not actually remain open after the open call
 *			if the data file requested was 'External' to the FPF.
 *		Warning: If the FPF was already open and this routine fails,
 *			then it will close the FPF.
 *		
 *
 * RETURN VALUE
 *		On success returns 'fpf_info'. On error returns NULL.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *	
*/
FPF_InfoType *FPF_Info (
	char			*filename,
	FPF_InfoType	*fpf_info
) 
{
	FPFHeaderType	fpfheader;

	if (!FPF_GetDir (filename, &fpfheader, NULL)) {
		return NULL;
	} else {
		fpf_info->NumFiles = fpfheader.NumFiles;
		fpf_info->NumDirFiles = fpfheader.NumDirFiles;
		return fpf_info;
	}
} /* FPF_Info */

/*********************************************************************
 *
 * FPF_findnext
 *
 * SYNOPSIS
 *		short FPF_findnext (
 *			struct FPF_tracker *fpftracker
 *		)
 *
 * PURPOSE
 *		To find subsequent files that match the 'pathname' given in 
 *		previous call to FPF_findfirst().  'fpftracker' is the same
 *		structure filled in by call to FPF_findfirst().  This structure
 *		contains necessary information for continuing the search.
 *		One FPF index for each call to FPF_findnext() will be returned
 *		until no more files are found in the FPF directory matching the
 *		'pathname'.
 *
 * INPUT
 *
 *
 * EFFECTS
 *		If a match is found, then the 'NameFound' field and 'FPF_IndexFound'
 *		field of 'fpftracker' are filled in with the name and index of
 *		the matched file.
 *
 *		This routine may be called on an FPF that is open or closed. 
 *		Note: An FPF does not actually remain open after the open call
 *			if the data file requested was 'External' to the FPF.
 *		Warning: If the FPF was already open and this routine fails,
 *			then it will close the FPF.
 *
 * RETURN VALUE
 *		On Success returns index of file in FPF (index >= 0). 
 *		If fails to match or gets error then returns -1 and errno is
 *		set to -1.
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		FPF_findfirst()
*/
short FPF_findnext (
	struct FPF_tracker *fpftracker
)
{
	FPFHeaderType	 fpfheader;
	FPFDirTableType	 dirtable;
	DirIndexesType	*dir_indexes;
	char			*filename_base;
	char			*filename;
	short			 fpf_index;

	/* Check if no more files to compare */
	if (fpftracker->CrntDirFileNum == -1)
	{
		return (errno = -1);
	}

	/* Read in FPF directory */
	if (!FPF_GetDir (fpftracker->FPF_Name, &fpfheader, &dirtable)) {
		return (errno = -1);
	}
	dir_indexes = dirtable.DirIndexes;
	filename_base = dirtable.FilenameBase;

	/* Search for matching filename */
	/* Linear Search from last found file */
	{
		short i;
		for (i = fpftracker->CrntDirFileNum; i < fpfheader.NumDirFiles; i++) 
		{
			filename = filename_base + dir_indexes[i].DirIndex;
			if (fstrpcmp (fpftracker->MatchPath, filename) == 0) {
				fpftracker->FPF_IndexFound = fpf_index =
					dir_indexes[i].FPF_Index;
				strcpy (fpftracker->NameFound, filename);
				FPF_FreeDir ();
				/* Set CrntDirFileNum to next file, or -1 if none left */
				fpftracker->CrntDirFileNum = 
					(++i >= fpfheader.NumDirFiles) ? (-1) : i;
				return fpf_index;
			}
		}
	}

//FREE_FAIL:

	/* Free FPF Directory */
	FPF_FreeDir ();

	return (errno = -1);

} /* FPF_findnext */

/*********************************************************************
 *
 * FPF_findfirst
 *
 * SYNOPSIS
 *		short	FPF_findfirst (
 *			char				*filename,
 *			char				*pathname,
 *			struct FPF_tracker	*fpftracker
 *		)
 *
 * PURPOSE
 *		To find the first file in an FPF directory matching 'path'.
 *
 * INPUT
 *		filename:	Name of FPF.
 *		pathname:	File name to find. File name part may contain '*'
 *					and '?' wild cards.
 *		fpftracker:	Pointer to tracking structure:
 *					stuct FPF_tracker {
 *						short	CrntDirFileNum;
 *						short	FPF_IndexFound;
 *						char	NameFound;
 *						char	MatchPath[EIO_MAXPATH];
 *						char	FPF_Name[EIO_MAXPATH];
 *					}
 *
 * EFFECTS
 *		Fills *fpftracker with information to be used later by 
 *		FPF_findnext().
 *	
 *		If a match is found, then the 'NameFound' field and 'FPF_IndexFound'
 *		field of 'fpftracker' are filled in with the name and index of
 *		the matched file.
 *
 *		This routine may be called on an FPF that is open or closed. 
 *		Note: An FPF does not actually remain open after the open call
 *			if the data file requested was 'External' to the FPF.
 *		Warning: If the FPF was already open and this routine fails,
 *			then it will close the FPF.
 *
 * RETURN VALUE
 *		On Success returns index of file in FPF (index >= 0). 
 *		If fails to match or gets error then returns -1 and errno is
 *		set to -1.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		FPF_findnext()
 *
*/
short	FPF_findfirst (
	char				*filename,
	char				*pathname,
	struct FPF_tracker	*fpftracker
)
{
	if (!filename || !pathname)
		return (errno = -1);

	strcpy (fpftracker->FPF_Name, filename);
	strcpy (fpftracker->MatchPath, pathname);
	fpftracker->CrntDirFileNum = 0;
	
	return FPF_findnext (fpftracker);
} /* FPF_findfirst */

